/**
 * Title:        StanfordMaxEnt<p>
 * Description:  A Maximum Entropy Toolkit<p>
 * Copyright:    Copyright (c) Kristina Toutanova<p>
 * Company:      Stanford University<p>
 */

package edu.stanford.nlp.tagger.maxent;

import java.io.Serializable;

/**
 * This class serves as the base class for classes which extract relevant
 * information from a history to give it to the features. Every feature has
 * an associated extractor or may be more.  GlobalHolder keeps all the
 * extractors; two histories are considered equal if all extractors return
 * equal values for them.  The main functionality of the Extractors is
 * provided by the method extract which takes a History as an argument.
 * The Extractor looks at the history and takes out something important for
 * the features - e.g. specific words and tags at specific positions or
 * some function of the History. The histories x are effectively vectors
 * of values, with each dimension being the output of some extractor.
 *
 * @author Kristina Toutanova
 * @version 1.0
 */
public class Extractor implements Serializable {

  static final String zeroSt = "0";

  private int position;
  boolean isTag;

  public Extractor() {
  }

  public boolean Precondition(String tag) {
    return true;
  }


  /**
   * @return the number of positions to the left the extractor looks at (only tags, because words are fixed.)
   */
  public int leftContext() {
    if (isTag) {
      if (position < 0) {
        return -position;
      }
    }

    return 0;
  }


  /**
   * @return the number of positions to the right the etxractor looks at (only tags, because words are fixed.)
   */
  public int rightContext() {
    if (isTag) {
      if (position > 0) {
        return position;
      }
    }

    return 0;
  }


  // CDM May 2007: This feature is currently never used. Maybe we should
  // change things so it is, and each feature template has a threshold, but
  // need to then work out what a TaggerFeature is and whether we should still
  // be using one of those to index with.
  // At present real threshold check happens in TaggerExperiments with
  // the populated(int, int) method. 
  public boolean isPopulated(TaggerFeature f) {
    return (f.indexedValues.length > GlobalHolder.minFeatureThresh);
  }

  /**
   * This constructor creates an extractor which extracts either the tag or
   * the word from position position in the history.
   *
   * @param position The position of the thing to be extracted. This is
   *                 relative to the current word. For example, position 0
   *                 will be the current word, -1 will be
   *                 the word before +1 will be the word after, etc.
   * @param isTag    If true this means that the POS tag is extracted from
   *                 position, otherwise the word is extracted.
   */
  public Extractor(int position, boolean isTag) {
    this.position = position;
    this.isTag = isTag;
  }


  String extract(History h) {
    return extract(h, GlobalHolder.pairs);
  }


  String extractLV(History h) {
    // should extract last verbal word and also the current word
    int start = h.start;
    String lastverb = "NA";
    int current = h.current;
    int index = current - 1;
    while (index >= start) {
      String tag = GlobalHolder.pairs.get(index, true);
      if (tag.startsWith("VB")) {
        lastverb = GlobalHolder.pairs.get(index, false);
        break;
      }
      if (tag.startsWith(",")) {
        break;
      }
      index--;
    }
    return lastverb;
  }


  String extract(History h, PairsHolder pH) {
    return pH.get(h, position, isTag);
  }


  String extract(History h, int bound) {
    return extract(h);
  }

  public String toString() {
    String cl = getClass().getName();
    int ind = cl.lastIndexOf(".");
    return cl.substring(ind + 1) + "(" + position + "," + (isTag ? "tag" : "word") + ")";
  }

  private static final long serialVersionUID = -4694133872973560083L;

}
