//ExtractorFrames -- StanfordMaxEnt, A Maximum Entropy Toolkit
//Copyright (c) 2002-2008 Leland Stanford Junior University


//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or (at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

//For more information, bug reports, fixes, contact:
//Christopher Manning
//Dept of Computer Science, Gates 1A
//Stanford CA 94305-9010
//USA
//    Support/Questions: java-nlp-user@lists.stanford.edu
//    Licensing: java-nlp-support@lists.stanford.edu
//http://www-nlp.stanford.edu/software/tagger.shtml


package edu.stanford.nlp.tagger.maxent;


/**
 * The static eFrames contains
 * an array of all Extractors that are used to define
 * features. This is an important class for the tagger.
 * If you want to add a new Extractor, you should add it to the
 * array eFrames.
 *
 * @author Kristina Toutanova
 * @version 1.0
 */
public class ExtractorFrames {

  static final Extractor cWord = new Extractor(0, false);
  static final Extractor cTag = new Extractor(0, true);
  static final Extractor prevTag = new Extractor(-1, true);
  static final Extractor prevTwoTag =new Extractor(-2,true);
  static final Extractor prevTwoTags = new ExtractorPrevTwoTags();
  static final Extractor prevThreeTags = new ExtractorPrevThreeTags();
  static final Extractor nextTwoTag =new Extractor(2,true);
  static final Extractor nextTwoTags = new ExtractorNextTwoTags();
  static final Extractor nextThreeTags = new ExtractorNextThreeTags();
  static final Extractor prevNextTag = new ExtractorPrevTagNextTag();
  static final Extractor prevNextTagWord = new ExtractorPrevTagNextTagWord();
  static final Extractor prevTagWord = new ExtractorPrevTagWord();
  static final Extractor nextTagWord = new ExtractorNextTagWord();
  static final Extractor nextWord = new Extractor(1, false);
  static final Extractor prevWord = new Extractor(-1, false);
  static final Extractor prevWord2 = new Extractor(-2,false);
  static final Extractor nextWord2 = new Extractor(2,false);

  static final Extractor lastVerb = new ExtractorLastVerb();
  static final Extractor lastVerbVB = new ExtractorVerbal_VB();
  static final Extractor lastVerbVBN = new ExtractorVerbal_VBN();
  static final Extractor lastVerbThat = new ExtractorLastVerbalTC();
  static final Extractor nextNNC = new ExtractorFollowingNN();
  static final Extractor lastVerbPT = new ExtractorParticlesChris();
  static final Extractor lastVerbDist = new ExtractorLVdist();
  static final Extractor nextTag = new Extractor(1, true);
  static final Extractor nextWordClass = new ExtractorFollowingWClass(1);
  static final Extractor nextNextWordClass = new ExtractorFollowingWClass(2);
  static final Extractor next2WordClass = new ExtractorFollowing2WClass();
  static final Extractor firstTaggerOutput = new ExtractorOutputTag(0, 0);
  static final Extractor secondTaggerOutput = new ExtractorOutputTag(1, 0);
  static final Extractor thirdTaggerOutput = new ExtractorOutputTag(2, 0);
  static final Extractor firstTaggernTag = new ExtractorOutputTag(0, 1);
  static final Extractor firstTaggern2Tags = new ExtractorNextTwoTagsTagger(0);
  static final Extractor allTaggerOutputs = new ExtractorAllTaggerOutputs();
  static final Extractor cWordNextWord = new ExtractorCWordNextWord();
  static final Extractor cWordPrevWord = new ExtractorCWordPrevWord();
  static final Extractor cWordLowerCase = new ExtractorCWordLowerCase();
  static final Extractor cWordCapCase = new ExtractorCWordCapCase();


  // this config was used in the best NAACL 2003 cyclic dependency tagger
  static final Extractor[] eFrames_bidirectional = {cWord,prevWord,nextWord,prevTag,
      nextTag,prevTwoTags,nextTwoTags,prevNextTag,prevTagWord,nextTagWord,
      cWordPrevWord,cWordNextWord};

  // this is a simple trigram CMM tagger (similar to the EMNLP 2000 tagger)
  static final Extractor[] eFrames_left3words = {cWord, prevWord, nextWord, prevTag, prevTwoTags};

  // like a simple trigram CMM tagger (similar to the EMNLP 2000 tagger) with more word context
  static final Extractor[] eFrames_left5words = {cWord, prevWord, nextWord, prevTag, prevTwoTags,
                                                  prevWord2, nextWord2};

  // like best NAACL 2003 cyclic dependency tagger but adds w_{-2}, w_{+2}
  static final Extractor[] eFrames_bidirectional5words = {cWord, prevWord, nextWord,
      prevWord2, nextWord2, prevTag, nextTag, prevTwoTags, nextTwoTags, prevNextTag,
      prevTagWord, nextTagWord, cWordPrevWord, cWordNextWord};

  /** This is the non-sequence model, 1 word context tagger */
  static final Extractor[] eFrames_1word = { cWord };

  /** This is the non-sequence model, 3 words context tagger */
  static final Extractor[] eFrames_3words = { cWord, prevWord, nextWord };

  /** This is the non-sequence model, 5 words context tagger */
  static final Extractor[] eFrames_5words = { cWord, prevWord, prevWord2, nextWord, nextWord2 };

  /** This is the non-sequence model, 7 words context tagger */
  static final Extractor[] eFrames_7words = { cWord, prevWord, prevWord2, new Extractor(-3, false), nextWord, nextWord2, new Extractor(3, false) };


  // features for 2005 SIGHAN tagger
  static final Extractor[] eFrames_sighan2005 = { cWord, prevWord, prevWord2, nextWord, nextWord2, prevTag, prevTwoTag, prevTwoTags };

  // features for a not-language-particular CMM tagger
  static final Extractor[] eFrames_generic ={ cWord, prevWord, nextWord,
      prevTag, prevTwoTags, prevTagWord, cWordPrevWord };


  // features for a german-language- CMM tagger
  static final Extractor[] eFrames_german ={ cWord, prevWord, nextWord, nextTag,
      prevTag, prevTwoTags, prevTagWord, cWordPrevWord };

  /**
   * This class is not meant to be instantiated.
   */
  private ExtractorFrames() {
  }

}


/**
 * the current word in lower-cased version
 * if the current word does not start with an alpha-numeric character, return 0
 */
@SuppressWarnings("serial")
class ExtractorCWordLowerCase extends Extractor {

  String extract(History h, PairsHolder pH) {
    return ExtractorFrames.cWord.extract(h, pH).toLowerCase();
  }

}


/**
 * conjunction of current word and last verb
 */
@SuppressWarnings("serial")
class ExtractorCWordLastVerb extends Extractor {
  private final static String excl = "!";

  String extract(History h, PairsHolder pH) {
    String cw = ExtractorFrames.cWord.extract(h, pH);
    String lastVer = ExtractorFrames.lastVerb.extract(h, pH);
    return cw + excl + lastVer;
  }

}


/**
 * the current word if it is capitalized, zero otherwise
 */
@SuppressWarnings("serial")
class ExtractorCWordCapCase extends Extractor {

  String extract(History h, PairsHolder pH) {
    String cw = ExtractorFrames.cWord.extract(h, pH);
    String lk = cw.toLowerCase();
    if (lk.equals(cw)) {
      return zeroSt;
    }
    return cw;
  }

}


/**
 * This extractor extracts the current and the next word in conjunction.
 */
@SuppressWarnings("serial")
class ExtractorCWordNextWord extends Extractor {

  private final static String excl = "!";

  public ExtractorCWordNextWord() {
  }

  String extract(History h, PairsHolder pH) {
    return pH.get(h, 0, false) + excl + pH.get(h, 1, false);
  }

}


/**
 * This extractor extracts the current and the next word and the next tag in conjunction.
 */
@SuppressWarnings("serial")
class ExtractorCWordNextWordTag extends Extractor {

  private final static String excl = "!";

  public ExtractorCWordNextWordTag() {
  }


  public int rightContext() {
    return 1;
  }

  String extract(History h, PairsHolder pH) {
    return pH.get(h, 0, false) + excl + pH.get(h, 1, true) + excl + pH.get(h, 1, false);
  }


}


/**
 * This extractor extracts the current and the previous word in conjunction.
 */
@SuppressWarnings("serial")
class ExtractorCWordPrevWord extends Extractor {

  private final static String excl = "!";

  public ExtractorCWordPrevWord() {
  }


  String extract(History h, PairsHolder pH) {
    return pH.get(h, -1, false) + excl + pH.get(h, 0, false);
  }

}


/**
 * This extractor extracts the current and the previous word in conjunction and also the
 * previous tag
 */
@SuppressWarnings("serial")
class ExtractorCWordPrevWordTag extends Extractor {

  private final static String excl = "!";

  public ExtractorCWordPrevWordTag() {
  }

  public int leftContext() {
    return 1;
  }


  String extract(History h, PairsHolder pH) {
    return pH.get(h, -1, false) + excl + pH.get(h, -1, true) + excl + pH.get(h, 0, false);
  }


}


/**
 * This extractor extracts the previous two tags.
 */
@SuppressWarnings("serial")
class ExtractorPrevTwoTags extends Extractor {

  private final static String excl = "!";

  public ExtractorPrevTwoTags() {
  }


  public int leftContext() {
    return 2;
  }


  String extract(History h, PairsHolder pH) {
    return pH.get(h, -1, true) + excl + pH.get(h, -2, true);
  }

}


/**
 * This extractor extracts the previous three tags.
 */
@SuppressWarnings("serial")
class ExtractorPrevThreeTags extends Extractor {

  private final static String excl = "!";

  public ExtractorPrevThreeTags() {
  }


  public int leftContext() {
    return 3;
  }


  String extract(History h, PairsHolder pH) {
    return pH.get(h, -1, true) + excl + pH.get(h, -2, true) + excl + pH.get(h, -3, true);
  }

}


/**
 * This extractor extracts the next two tags.
 */

@SuppressWarnings("serial")
class ExtractorNextTwoTags extends Extractor {

  private final static String excl = "!";

  public ExtractorNextTwoTags() {
  }


  public int rightContext() {
    return 2;
  }


  String extract(History h, PairsHolder pH) {
    return pH.get(h, 1, true) + excl + pH.get(h, 2, true);
  }

}


/**
 * This extractor extracts the next three tags.
 */
@SuppressWarnings("serial")
class ExtractorNextThreeTags extends Extractor {

  private final static String excl = "!";

  public ExtractorNextThreeTags() {
  }

  public int rightContext() {
    return 3;
  }


  String extract(History h, PairsHolder pH) {
    return pH.get(h, 1, true) + excl + pH.get(h, 2, true) + excl + pH.get(h, 3, true);
  }

}


/**
 * This extractor extracts the previous tag and the current word in conjunction.
 */
@SuppressWarnings("serial")
class ExtractorPrevTagWord extends Extractor {

  private final static String excl = "!";

  public ExtractorPrevTagWord() {
  }


  public int leftContext() {
    return 1;
  }

  String extract(History h, PairsHolder pH) {
    return pH.get(h, -1, true) + excl + pH.get(h, 0, false);
  }

}


/**
 * This extractor extracts the previous tag , next tag, and the current word in conjunction.
 */
@SuppressWarnings("serial")
class ExtractorPrevTagNextTagWord extends Extractor {

  private final static String excl = "!";

  public ExtractorPrevTagNextTagWord() {
  }


  public int leftContext() {
    return 1;
  }

  public int rightContext() {
    return 1;
  }


  String extract(History h, PairsHolder pH) {
    return pH.get(h, -1, true) + excl + pH.get(h, 0, false) + pH.get(h, 1, true);
  }

}


/**
 * This extractor extracts the previous tag , next tag in conjunction.
 */
@SuppressWarnings("serial")
class ExtractorPrevTagNextTag extends Extractor {

  private final static String excl = "!";

  public ExtractorPrevTagNextTag() {
  }


  public int leftContext() {
    return 1;
  }

  public int rightContext() {
    return 1;
  }


  String extract(History h, PairsHolder pH) {
    return pH.get(h, -1, true) + excl + pH.get(h, 1, true);
  }

}


/**
 * This extractor extracts the next tag and the current word in conjunction.
 */
@SuppressWarnings("serial")
class ExtractorNextTagWord extends Extractor {

  private final static String excl = "!";

  public ExtractorNextTagWord() {
  }


  public int rightContext() {
    return 1;
  }


  String extract(History h, PairsHolder pH) {
    return pH.get(h, 1, true) + excl + pH.get(h, 0, false);
  }

}


/**
 * This extractor extracts the next two tags of some tagger.
 */
@SuppressWarnings("serial")
class ExtractorNextTwoTagsTagger extends Extractor {

  Extractor e1;
  Extractor e2;
  int num;


  public int rightContext() {
    return 2;
  }

  public ExtractorNextTwoTagsTagger(int num) {
    this.num = num;
    e1 = new ExtractorOutputTag(num, 1);
    e2 = new ExtractorOutputTag(num, 2);
  }


  String extract(History h, PairsHolder pH) {
    return e1.extract(h, pH) + e2.extract(h, pH);
  }

}




