//ExtractorFramesRare -- StanfordMaxEnt, A Maximum Entropy Toolkit
//Copyright (c) 2002-2008 Leland Stanford Junior University


//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or (at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

//For more information, bug reports, fixes, contact:
//Christopher Manning
//Dept of Computer Science, Gates 1A
//Stanford CA 94305-9010
//USA
//    Support/Questions: java-nlp-user@lists.stanford.edu
//    Licensing: java-nlp-support@lists.stanford.edu
//http://www-nlp.stanford.edu/software/tagger.shtml

package edu.stanford.nlp.tagger.maxent;

import edu.stanford.nlp.maxent.BinaryFeature;
import edu.stanford.nlp.process.WordShapeClassifier;

import java.util.HashSet;
import java.util.Set;

/**
 * Provides arrays of ExtractorFrames for rare words.  This file
 * also defines all the rare word extractors as non-public classes.
 * This file is used simply as a resource to initialize extractors stored
 * in Extractors class instances.
 *
 * @author Kristina Toutanova
 * @author Christopher Manning
 * @version 2.0
 */
public class ExtractorFramesRare {

  /**
   * Last 1-4 characters of word
   */
  private static Extractor cWordSuff1 = new ExtractorCWordSuff(1);
  private static Extractor cWordSuff2 = new ExtractorCWordSuff(2);
  private static Extractor cWordSuff3 = new ExtractorCWordSuff(3);
  private static Extractor cWordSuff4 = new ExtractorCWordSuff(4);

  /**
   * First 1-4 characters of word
   */
  private static Extractor cWordPref1 = new ExtractorCWordPref(1);
  private static Extractor cWordPref2 = new ExtractorCWordPref(2);
  private static Extractor cWordPref3 = new ExtractorCWordPref(3);
  private static Extractor cWordPref4 = new ExtractorCWordPref(4);

  /**
   * "1" iff word contains 1 or more upper case characters (somewhere)
   */
  private static Extractor cWordUppCase = new ExtractorUCase();

  /**
   * "1" iff word contains 1 or more digit characters (somewhere)
   */
  private static Extractor cWordNumber = new ExtractorCNumber();

  /**
   * "1" iff word contains 1 or more dash characters (somewhere)
   */
  private static Extractor cWordDash = new ExtractorDash();

  /**
   * "0" if not 1st word of sentence or not upper case, or lowercased version
   * not in dictionary.  Else first tag of word lowercased.
   */
  private static Extractor cWordStartUCase = new ExtractorStartSentenceCap();

  /**
   * "0" if first word of sentence or not first letter uppercase or if
   * lowercase version isn't in dictionary.  Otherwise first tag of lowercase
   * equivalent.
   */
  private static Extractor cWordMidUCase = new ExtractorMidSentenceCapC();

  /**
   * "0" if first letter isn't uppercase or if
   * lowercase version isn't in dictionary.  Otherwise first tag of lowercase
   * equivalent.
   */
  // private static Extractor cUCaseLCase = new ExtractorCapC();

  /**
   * "1" if not first word of sentence and _some_ letter is uppercase
   */
  private static Extractor cMidSentence = new ExtractorMidSentenceCap();

  /**
   * "1" if token has no lower case letters
   */
  private static Extractor cAllCap = new ExtractorAllCap();

  /**
   * "1" if token has only upper case letters
   */
  private static Extractor cAllCapitalized = new ExtractorAllCapitalized();

  /**
   * "1" if capitalized and one of following 3 words is Inc., Co., or Corp.
   */
  private static Extractor cCompany = new CompanyNameDetector();

  /**
   * "1" if a plural acronym: capital letters followed by 's'
   */
  // private static Extractor cPluralAcronym = new PluralAcronymDetector();

  /**
   * "1" if word contains letter, digit, and dash, in any position and case
   */
  private static Extractor cLetterDigitDash = new ExtractorLetterDigitDash();

  /**
   * "1" if word contains uppercase letter, digit, and dash
   */
  private static Extractor cUpperDigitDash = new ExtractorUpperDigitDash();

  /**
   * Distance to lowercase word.  Used by another extractor....
   */
  private static Extractor cCapDist = new ExtractorCapDistLC();




  private static final Extractor[] eFrames_naacl2003 = { cWordUppCase, cWordNumber, cWordDash, cWordSuff1, cWordSuff2, cWordSuff3, cWordSuff4, cAllCap, cMidSentence, cWordStartUCase, cWordMidUCase, cWordPref1, cWordPref2, cWordPref3, cWordPref4, new ExtractorCWordPref(5), new ExtractorCWordPref(6), new ExtractorCWordPref(7), new ExtractorCWordPref(8), new ExtractorCWordPref(9), new ExtractorCWordPref(10), new ExtractorCWordSuff(5), new ExtractorCWordSuff(6), new ExtractorCWordSuff(7), new ExtractorCWordSuff(8), new ExtractorCWordSuff(9), new ExtractorCWordSuff(10), cLetterDigitDash, cCompany, cAllCapitalized, cUpperDigitDash};

  private static final Extractor[] eFrames_naacl2003wordShape3 = { cWordUppCase, cWordNumber, cWordDash, cWordSuff1, cWordSuff2, cWordSuff3, cWordSuff4, cAllCap, cMidSentence, cWordStartUCase, cWordMidUCase, cWordPref1, cWordPref2, cWordPref3, cWordPref4, new ExtractorCWordPref(5), new ExtractorCWordPref(6), new ExtractorCWordPref(7), new ExtractorCWordPref(8), new ExtractorCWordPref(9), new ExtractorCWordPref(10), new ExtractorCWordSuff(5), new ExtractorCWordSuff(6), new ExtractorCWordSuff(7), new ExtractorCWordSuff(8), new ExtractorCWordSuff(9), new ExtractorCWordSuff(10), cLetterDigitDash, cCompany, cAllCapitalized, cUpperDigitDash, new ExtractorWordShapeClassifier(0, "chris2"),new ExtractorWordShapeClassifier(-1, "chris2"),new ExtractorWordShapeClassifier(1, "chris2")};

  private static final Extractor[] eFrames_generic = { cWordSuff1, cWordSuff2, cWordSuff3, cWordSuff4, cWordPref1, cWordPref2, cWordPref3, cWordPref4, new ExtractorCWordPref(5), new ExtractorCWordPref(6), new ExtractorCWordSuff(5), new ExtractorCWordSuff(6) };


  private static final Extractor[] eFrames_sighan2005 = { cWordNumber, cWordSuff1 , cWordSuff2, cWordSuff3, cWordSuff4 , cWordPref1, cWordPref2, cWordPref3, cWordPref4};

  /** bio extractor frames are still under-development; currently better to just use naacl */
  private static final Extractor[] eFrames_bio = { cWordDash, cWordSuff1, cWordSuff2, cWordSuff3, cWordSuff4, cWordPref1, cWordPref2, cWordPref3, cWordPref4, new ExtractorCWordPref(5), new ExtractorCWordPref(6), new ExtractorCWordSuff(5), new ExtractorCWordSuff(6) };


  /** Get an array of Extractor identified by a name.
   *
   *  @param identifier Describes a set of extractors for rare word features
   *  @return A set of extractors for rare word features
   */
  public static Extractor[] getExtractorFramesRare(String identifier, TaggerConfig config) {
    Extractor[] eFrames = eFrames_generic;
    if ("naacl2003".equals(identifier)) {
      eFrames = eFrames_naacl2003;
    } else if ("naacl2003conjunctions".equals(identifier)) {
      eFrames = addConjunctions(eFrames_naacl2003);
    } else if ("naacl2003wordshape3".equals(identifier)) {
      eFrames = eFrames_naacl2003wordShape3;
    } else if ("sighan2005".equals(identifier)) {
      eFrames = addSighanChineseFeatures(eFrames_sighan2005, config.getUseChineseDict());
    } else if ("generic".equals(identifier)) {
      eFrames = eFrames_generic;
    } else {
      System.err.println("Unrecognized ExtractorFramesRare identifier: " + identifier + ", using generic.");
      eFrames = eFrames_generic;
    }
    if(config != null && config.getUseDistSim()) {
      eFrames = addDistSimExtractor(eFrames, config.getDistSimPath());
    }

    return eFrames;
  }


  /**
   * This adds the conjunctions of various features in the rare words features.
   * @param extIn A base array of extractors
   * @return An expanded array
   */
  private static Extractor[] addConjunctions(Extractor[] extIn) {
    Extractor[] newW = new Extractor[24];
    //add them manually ....
    newW[0] = new ExtractorsConjunction(cWordUppCase, cWordSuff1);
    newW[1] = new ExtractorsConjunction(cWordUppCase, cWordSuff2);
    newW[2] = new ExtractorsConjunction(cWordUppCase, cWordSuff3);
    newW[3] = new ExtractorsConjunction(cWordUppCase, cWordSuff4);

    newW[4] = new ExtractorsConjunction(cAllCap, cWordSuff1);
    newW[5] = new ExtractorsConjunction(cAllCap, cWordSuff2);
    newW[6] = new ExtractorsConjunction(cAllCap, cWordSuff3);
    newW[7] = new ExtractorsConjunction(cAllCap, cWordSuff4);

    newW[8] = new ExtractorsConjunction(cMidSentence, cWordSuff1);
    newW[9] = new ExtractorsConjunction(cMidSentence, cWordSuff2);
    newW[10] = new ExtractorsConjunction(cMidSentence, cWordSuff3);
    newW[11] = new ExtractorsConjunction(cMidSentence, cWordSuff4);

    newW[12] = new ExtractorsConjunction(cWordStartUCase, cWordSuff1);
    newW[13] = new ExtractorsConjunction(cWordStartUCase, cWordSuff2);
    newW[14] = new ExtractorsConjunction(cWordStartUCase, cWordSuff3);
    newW[15] = new ExtractorsConjunction(cWordStartUCase, cWordSuff4);

    newW[16] = new ExtractorsConjunction(cWordMidUCase, cWordSuff1);
    newW[17] = new ExtractorsConjunction(cWordMidUCase, cWordSuff2);
    newW[18] = new ExtractorsConjunction(cWordMidUCase, cWordSuff3);
    newW[19] = new ExtractorsConjunction(cWordMidUCase, cWordSuff4);

    newW[20] = new ExtractorsConjunction(cCapDist, cWordSuff1);
    newW[21] = new ExtractorsConjunction(cCapDist, cWordSuff2);
    newW[22] = new ExtractorsConjunction(cCapDist, cWordSuff3);
    newW[23] = new ExtractorsConjunction(cCapDist, cWordSuff4);

    Extractor[] eFramesNew = new Extractor[extIn.length + newW.length];

    System.arraycopy(extIn, 0, eFramesNew, 0, extIn.length);
    System.arraycopy(newW, 0, eFramesNew, extIn.length, newW.length);
    return eFramesNew;
  }

  /**
   * Adds a distsim extractor, with distsim info at the appropriate path, to the given extractor array
   * @param extIn
   * @return
   */
  private static Extractor[] addDistSimExtractor(Extractor[] extIn, String distSimPath) {
    Extractor distsim = new ExtractorDistsim(distSimPath);
    Extractor[] eFramesNew = new Extractor[extIn.length + 1];
    System.arraycopy(extIn, 0, eFramesNew, 0, extIn.length);
    eFramesNew[eFramesNew.length - 1]  = distsim;
    return eFramesNew;
  }


  private static Extractor[] addLCTagFeatures(Extractor[] extIn) {
    Extractor[] newE = new Extractor[GlobalHolder.tags.getSize()];
    for (int i = 0; i < GlobalHolder.tags.getSize(); i++) {
      String tag = GlobalHolder.tags.getTag(i);
      newE[i] = new ExtractorCapLCSeen(tag);
    }

    Extractor[] eFramesNew = new Extractor[extIn.length + newE.length];

    System.arraycopy(extIn, 0, eFramesNew, 0, extIn.length);
    System.arraycopy(newE, 0, eFramesNew, extIn.length, newE.length);

    return eFramesNew;
  }


  public ExtractorFramesRare() {

    System.out.println("calling init for extractorframesrare ");
    System.exit(0);

    /*
    ArrayList<Extractor> v = new ArrayList<Extractor>();
    GlobalHolder.ySize = GlobalHolder.tags.getSize();
    for (int i = 1; i < 5; i++) {
      for (int y = 0; y < GlobalHolder.tags.getSize(); y++) {
        if (!GlobalHolder.tags.isClosed(GlobalHolder.tags.getTag(y))) {
          ExtractorMorpho extr = new ExtractorMorpho(i, y);
          v.add(extr);
        }// if open
      }
    }// for i

    for (int y = 0; y < GlobalHolder.ySize; y++) {
      for (int y1 = 0; y1 < GlobalHolder.ySize; y1++) {
        if (!GlobalHolder.tags.isClosed(GlobalHolder.tags.getTag(y)) && (!GlobalHolder.tags.isClosed(GlobalHolder.tags.getTag(y)))) {
          ExtractorMorpho extr = new ExtractorMorpho(5, y, y1);
          v.add(extr);
        }// if open
      }
    }
    int vSize = v.size();
    Extractor[] eFramestemp = new Extractor[eFrames.length + vSize];
    System.arraycopy(eFrames, 0, eFramestemp, 0, eFrames.length);
    for (int i = 0; i < vSize; i++) {
      eFramestemp[i + eFrames.length] = v.get(i);
    }
    eFrames = eFramestemp;
     */
  }


  /** Adds features for 2005 SIGHAN tagger.
   */
  public static Extractor[] addSighanChineseFeatures(Extractor[] extIn, boolean useChineseDict) {
    Extractor[] extOut = extIn;
    for (int i = -2; i <= 2; i++) {
      if(useChineseDict) {
        extOut = addCTBPreFeatures(extOut, i);
        extOut = addCTBSufFeatures(extOut, i);
        extOut = addCTBUnkDictFeatures(extOut, i);
        extOut = addASBCUnkFeatures(extOut, i);
      }
    }
    return extOut;
  }


  public static Extractor[] addCTBPreFeatures(Extractor[] extIn, int n) {
    String[] tagsets = new String[]{"AD", "AS", "BA", "CC", "CD", "CS", "DEC", "DEG", "DER", "DEV", "DT", "ETC", "FW", "IJ", "JJ", "LB", "LC", "M", "MSP", "NN", "NP", "NR", "NT", "OD", "P", "PN", "PU", "SB", "SP", "VA", "VC", "VE", "VV" };
    Extractor[] newW = new Extractor[tagsets.length];
    for(int k=0;k<tagsets.length;k++){
      newW[k] = new ctbPreDetector(tagsets[k], n);
    }
    Extractor[] eFramesNew = new Extractor[extIn.length+newW.length];

    System.arraycopy(extIn, 0, eFramesNew, 0, extIn.length);
    System.arraycopy(newW, 0, eFramesNew, extIn.length, newW.length);

    return eFramesNew;
  } // end addCTBPreFeatures


  public static Extractor[] addCTBSufFeatures(Extractor[] extIn, int n) {
    String[] tagsets = new String[]{"AD", "AS", "BA", "CC", "CD", "CS", "DEC", "DEG", "DER", "DEV", "DT", "ETC", "FW", "IJ", "JJ", "LB", "LC", "M", "MSP", "NN", "NP", "NR", "NT", "OD", "P", "PN", "PU", "SB", "SP", "VA", "VC", "VE", "VV" };
    Extractor[] newW=new Extractor[tagsets.length];
    for(int k=0;k<tagsets.length;k++){
      newW[k] = new ctbSufDetector(tagsets[k], n);
    }
    Extractor[] eFramesNew=new Extractor[extIn.length+newW.length];

    System.arraycopy(extIn, 0, eFramesNew, 0, extIn.length);
    System.arraycopy(newW, 0, eFramesNew, extIn.length, newW.length);

    return eFramesNew;
  } //end


  public static Extractor[] addASBCPreFeatures(Extractor[] extIn, int n) {

    String[] tagsets = new String[]{"A", "Caa", "Cab", "Cba", "Cbb", "D", "DE", "DK", "Da", "Dd", "De", "Des", "Dfa", "Dfb", "Di", "Dk", "FW", "I", " Na", "Nb", " Nc", "Ncb", "Ncd", " Nd", "Neaq", "Nep", "Neqa", "Neqb", "Nes", "Neu", "Nf", "Ng", "Nh", "P", "PU", "SHI", "T", "VA", "VAC", "VB", "VC", "VCL", "VD", "VE", "VF", "VG", "VH", "VHC", "VI", "VJ", "VK", "VL", "V_2" };
    Extractor[] newW=new Extractor[tagsets.length];
    for(int k=0;k<tagsets.length;k++){
      newW[k] = new ASBCPreDetector(tagsets[k], n);
    }
    Extractor[] eFramesNew=new Extractor[extIn.length+newW.length];

    System.arraycopy(extIn, 0, eFramesNew, 0, extIn.length);
    System.arraycopy(newW, 0, eFramesNew, extIn.length, newW.length);

    return eFramesNew;
  } //end


  public static Extractor[] addASBCsufFeatures(Extractor[] extIn, int n) {

    String[] tagsets = new String[]{"A", "Caa", "Cab", "Cba", "Cbb", "D", "DE", "DK", "Da", "Dd", "De", "Des", "Dfa", "Dfb", "Di", "Dk", "FW", "I", " Na", "Nb", " Nc", "Ncb", "Ncd", " Nd", "Neaq", "Nep", "Neqa", "Neqb", "Nes", "Neu", "Nf", "Ng", "Nh", "P", "PU", "SHI", "T", "VA", "VAC", "VB", "VC", "VCL", "VD", "VE", "VF", "VG", "VH", "VHC", "VI", "VJ", "VK", "VL", "V_2"  };
    Extractor[] newW=new Extractor[tagsets.length];
    for(int k=0;k<tagsets.length;k++){
      newW[k] = new ASBCSufDetector(tagsets[k], n);
    }
    Extractor[] eFramesNew=new Extractor[extIn.length+newW.length];

    System.arraycopy(extIn, 0, eFramesNew, 0, extIn.length);
    System.arraycopy(newW, 0, eFramesNew, extIn.length, newW.length);

    return eFramesNew;
  } //end


  public static Extractor[] addASBCUnkFeatures(Extractor[] extIn, int n) {

    String[] tagsets = new String[]{"A", "Caa", "Cab", "Cba", "Cbb", "D", "DE", "DK", "Da", "Dd", "De", "Des", "Dfa", "Dfb", "Di", "Dk", "FW", "I", " Na", "Nb", " Nc", "Ncb", "Ncd", " Nd", "Neaq", "Nep", "Neqa", "Neqb", "Nes", "Neu", "Nf", "Ng", "Nh", "P", "PU", "SHI", "T", "VA", "VAC", "VB", "VC", "VCL", "VD", "VE", "VF", "VG", "VH", "VHC", "VI", "VJ", "VK", "VL", "V_2"  };

    Extractor[] newW=new Extractor[tagsets.length];
    for(int k=0;k<tagsets.length;k++){
      newW[k] = new ASBCunkDetector(tagsets[k], n);
    }
    Extractor[] eFramesNew=new Extractor[extIn.length+newW.length];

    System.arraycopy(extIn, 0, eFramesNew, 0, extIn.length);
    System.arraycopy(newW, 0, eFramesNew, extIn.length, newW.length);

    return eFramesNew;
  } //end


  public static Extractor[] addCTBUnkDictFeatures(Extractor[] extIn, int n) {

    String[] tagsets = new String[]{"A", "Caa", "Cab", "Cba", "Cbb", "D", "DE", "DK", "Da", "Dd", "De", "Des", "Dfa", "Dfb", "Di", "Dk", "FW", "I", " Na", "Nb", " Nc", "Ncb", "Ncd", " Nd", "Neaq", "Nep", "Neqa", "Neqb", "Nes", "Neu", "Nf", "Ng", "Nh", "P", "PU", "SHI", "T", "VA", "VAC", "VB", "VC", "VCL", "VD", "VE", "VF", "VG", "VH", "VHC", "VI", "VJ", "VK", "VL", "V_2"  };

    Extractor[] newW=new Extractor[tagsets.length];
    for(int k=0;k<tagsets.length;k++){
      newW[k] = new CTBunkDictDetector(tagsets[k], n);
    }
    Extractor[] eFramesNew=new Extractor[extIn.length+newW.length];

    System.arraycopy(extIn, 0, eFramesNew, 0, extIn.length);
    System.arraycopy(newW, 0, eFramesNew, extIn.length, newW.length);

    return eFramesNew;
  }

} // end class ExtractorFramesRare


/**
 * Superclass for rare word feature frames.  Provides some common functions.
 * Designed to be extended.
 */
class RareExtractor extends Extractor {

  public static final String naTag = "NA";

  public boolean isPopulated(BinaryFeature f) {
    return f.indexedValues.length > GlobalHolder.rareWordMinFeatureThresh;
  }

  public static boolean startsUpperCase(String s) {
    if (s == null || s.length() == 0) {
      return false;
    }
    char ch = s.charAt(0);
    return Character.isUpperCase(ch);
  }

  /**
   * a string is lowercase if it starts with a lowercase letter
   * such as one from a to z.
   * Should we include numbers?
   * @param s The String to check
   * @return If its first character is lower case
   */
  public static boolean startsLowerCase(String s) {
    if (s == null) {
      return false;
    }
    char ch = s.charAt(0);
    return Character.isLowerCase(ch);
  }

  public static boolean containsDash(String s) {
    return s != null && s.indexOf("-") >= 0;
  }

  public static boolean containsNumber(String s) {
    if (s == null) {
      return false;
    }
    for (int i = 0, len = s.length(); i < len; i++) {
      if (Character.isDigit(s.charAt(i))) {
        return true;
      }
    }
    return false;
  }

  public static boolean containsLetter(String s) {
    if (s == null) {
      return false;
    }
    for (int i = 0, len = s.length(); i < len; i++) {
      if (Character.isLetter(s.charAt(i))) {
        return true;
      }
    }
    return false;
  }

  public static boolean containsUpperCase(String s) {
    if (s == null) {
      return false;
    }
    for (int i = 0, len = s.length(); i < len; i++) {
      if (Character.isUpperCase(s.charAt(i))) {
        return true;
      }
    }
    return false;
  }

  public static boolean allUpperCase(String s) {
    if (s == null) {
      return false;
    }
    for (int i = 0, len = s.length(); i < len; i++) {
      if (!Character.isUpperCase(s.charAt(i))) {
        return false;
      }
    }
    return true;
  }

  public static boolean noneLowerCase(String s) {
    if (s == null) {
      return false;
    }
    for (int i = 0, len = s.length(); i < len; i++) {
      if (Character.isLowerCase(s.charAt(i))) {
        return false;
      }
    }
    return true;
  }


  private static final long serialVersionUID = -7682607870855426599L;

} // end class RareExtractor


class CompanyNameDetector extends RareExtractor {

  private Set<String> companyNameEnds;

  public CompanyNameDetector() {
    companyNameEnds = new HashSet<String>();
    companyNameEnds.add("Company");
    companyNameEnds.add("COMPANY");
    companyNameEnds.add("Co.");
    companyNameEnds.add("Co");  // at end of sentence in PTB
    companyNameEnds.add("Cos.");
    companyNameEnds.add("CO.");
    companyNameEnds.add("COS.");
    companyNameEnds.add("Corporation");
    companyNameEnds.add("CORPORATION");
    companyNameEnds.add("Corp.");
    companyNameEnds.add("Corp"); // at end of sentence in PTB
    companyNameEnds.add("CORP.");
    companyNameEnds.add("Incorporated");
    companyNameEnds.add("INCORPORATED");
    companyNameEnds.add("Inc.");
    companyNameEnds.add("Inc"); // at end of sentence in PTB
    companyNameEnds.add("INC.");
    companyNameEnds.add("Association");
    companyNameEnds.add("ASSOCIATION");
    companyNameEnds.add("Assn");
    companyNameEnds.add("ASSN");
    companyNameEnds.add("Limited");
    companyNameEnds.add("LIMITED");
    companyNameEnds.add("Ltd.");
    companyNameEnds.add("LTD.");
    companyNameEnds.add("L.P.");
  }

  private boolean companyNameEnd(String s) {
    // the null test seemed to be needed when deserializing against old
    // tagger versions....
    return companyNameEnds != null && s != null && companyNameEnds.contains(s);
  }

  String extract(History h, PairsHolder pH) {
    String s = pH.get(h, 0, false);
    if (!startsUpperCase(s)) {
      return "0";
    }
    for (int i = 0; i <= 3; i++) {
      String s1 = pH.get(h, i, false);
      if (companyNameEnd(s1)) {
        return "1";
      }
    }
    return "0";
  }

  private static final long serialVersionUID = 21;

} // end class CompanyNameDetector


class ExtractorUCase extends RareExtractor {

  public ExtractorUCase() {
  }

  String extract(History h, PairsHolder pH) {
    String s = pH.get(h, 0, false);
    if (containsUpperCase(s)) {
      return "1";
    }
    return "0";
  }

  private static final long serialVersionUID = 22;

}


class ExtractorLetterDigitDash extends RareExtractor {

  public ExtractorLetterDigitDash() {
  }

  String extract(History h, PairsHolder pH) {
    String s = pH.get(h, 0, false);
    if (containsLetter(s) && containsDash(s) && containsNumber(s)) {
      return "1";
    }
    return "0";
  }

  private static final long serialVersionUID = 23;

}


class ExtractorUpperDigitDash extends RareExtractor {

  public ExtractorUpperDigitDash() {
  }

  String extract(History h, PairsHolder pH) {
    String s = pH.get(h, 0, false);
    if (containsUpperCase(s) && containsDash(s) && containsNumber(s)) {
      return "1";
    }
    return "0";
  }

  private static final long serialVersionUID = 33;

}


/**
 * creates features which are true if the current word is all caps
 * and the distance to the first lowercase word to the left is dist
 * the distance is 1 for adjacent, 2 for one across, 3 for ... and so on.
 * inifinity if no capitalized word (we hit the start of sentence or '')
 */
class ExtractorCapDistLC extends RareExtractor {

  boolean verbose = false;

  public boolean Precondition(String tag) {
    return true;
    //(tag.equals("NNP")||tag.equals("NNPS"));
  }

  public ExtractorCapDistLC() {
  }

  String extract(History h, PairsHolder pH) {
    String word = pH.get(h, 0, false);
    String ret;
    if (!startsUpperCase(word)) {
      if (verbose) {
        System.out.println("did not apply because not start with upper case");
      }
      return "0";
    }
    if (allUpperCase(word)) {
      ret = "all:";
    } else {
      ret = "start";
    }

    //now find the distance
    int current = -1;
    int distance = 1;

    while (true) {
      String prevWord = pH.get(h, current, false);
      if (startsLowerCase(prevWord)) {
        if (verbose) {
          System.out.println("returning " + (ret + current) + "for " + word + " " + prevWord);
        }
        return ret + distance;
      }
      if (prevWord.equals(naTag) || prevWord.equals("``")) {
        if (verbose) {
          System.out.println("returning " + ret + "infinity for " + word + " " + prevWord);
        }
        return ret + "infinity";
      }
      current--;
      distance++;
    }
  }

  private static final long serialVersionUID = 34;

}


/**
 * this feature applies when the word is capitalized
 * and the previous lower case is infinity
 * and the lower cased version of it has occured 2 or more times with tag t
 * false if the word was not seen
 * create features only for tags that are the same as the tag t
 */
class ExtractorCapLCSeen extends RareExtractor {

  String tag;
  int cutoff = 1;
  private Extractor cCapDist = new ExtractorCapDistLC();

  public ExtractorCapLCSeen(String tag) {
    this.tag = tag;
  }

  public boolean Precondition(String tag1) {
    return tag.equals(tag1);
  }

  String extract(History h, PairsHolder pH) {
    String res = cCapDist.extract(h);
    if (res.equals("0")) {
      return res;
    }
    //otherwise it is capitalized
    String word = ExtractorFrames.cWord.extract(h);
    if (GlobalHolder.dict.getCount(word, tag) > cutoff) {
      return res + tag;
    } else {
      return "0";
    }
  }

  private static final long serialVersionUID = 35;

}


class ExtractorMidSentenceCap extends RareExtractor {

  public ExtractorMidSentenceCap() {
  }

  String extract(History h, PairsHolder pH) {
    String prevTag = pH.get(h, -1, true);
    if(prevTag == null) { return "0"; } /* i'm not sure why this is
     * happening, and i really
     * don't want to spend a month
     * tracing it
     * down. -wmorgan */
    if (prevTag.equals(naTag)) {
      return "0";
    }
    String s = pH.get(h, 0, false);
    if (containsUpperCase(s)) {
      return "1";
    }
    return "0";
  }

  private static final long serialVersionUID = 24;

}


class ExtractorStartSentenceCap extends RareExtractor {

  public ExtractorStartSentenceCap() {
  }

  String extract(History h, PairsHolder pH) {
    String prevTag = pH.get(h, -1, true);
    if(prevTag == null) { return zeroSt; } /* i'm not sure why this is
     * happening, and i really
     * don't want to spend a month
     * tracing it
     * down. -wmorgan */

    if (!prevTag.equals(naTag)) {
      return zeroSt;
    }
    String s = pH.get(h, 0, false);
    if (startsUpperCase(s)) {
      String s1 = s.toLowerCase();
      if (GlobalHolder.dict.sum(s1) == 0) {
        return zeroSt;
      }
      return GlobalHolder.dict.getFirstTag(s1);
    }
    return zeroSt;
  }

  private static final long serialVersionUID = 25;

}


class ExtractorMidSentenceCapC extends RareExtractor {

  public ExtractorMidSentenceCapC() {
  }

  String extract(History h, PairsHolder pH) {
    String prevTag = pH.get(h, -1, true);
    if (prevTag == null) { return zeroSt; } /* i'm not sure why this is
     * happening, and i really
     * don't want to spend a month
     * tracing it
     * down. -wmorgan */
    if (prevTag.equals(naTag)) {
      return zeroSt;
    }
    String s = pH.get(h, 0, false);
    if (startsUpperCase(s)) {
      String s1 = s.toLowerCase();
      if (GlobalHolder.dict.sum(s1) == 0) {
        return zeroSt;
      }
      return GlobalHolder.dict.getFirstTag(s1);
    }
    return zeroSt;
  }

  private static final long serialVersionUID = 26L;

}


class ExtractorCapC extends RareExtractor {

  public ExtractorCapC() {
  }

  String extract(History h, PairsHolder pH) {
    String s = pH.get(h, 0, false);
    if (startsUpperCase(s)) {
      String s1 = s.toLowerCase();
      if (GlobalHolder.dict.sum(s1) == 0) {
        return zeroSt;
      }
      return GlobalHolder.dict.getFirstTag(s1);
    }
    return zeroSt;
  }

  private static final long serialVersionUID = 26;

}


class ExtractorAllCap extends RareExtractor {

  public ExtractorAllCap() {
  }

  String extract(History h, PairsHolder pH) {
    String s = pH.get(h, 0, false);
    if (noneLowerCase(s)) {
      return "1";
    }
    return "0";
  }

  private static final long serialVersionUID = 27;

}


class ExtractorAllCapitalized extends RareExtractor {

  public ExtractorAllCapitalized() {
  }

  String extract(History h, PairsHolder pH) {
    String s = pH.get(h, 0, false);
    if (allUpperCase(s)) {
      return "1";
    }
    return "0";
  }

  private static final long serialVersionUID = 32L;

}


class ExtractorCNumber extends RareExtractor {

  public ExtractorCNumber() {
  }

  String extract(History h, PairsHolder pH) {
    String s = pH.get(h, 0, false);
    if (containsNumber(s)) {
      return "1";
    }
    return "0";
  }

  private static final long serialVersionUID = 28;

}


class ExtractorDash extends RareExtractor {

  public ExtractorDash() {
  }

  String extract(History h, PairsHolder pH) {
    String s = pH.get(h, 0, false);
    if (containsDash(s)) {
      return "1";
    }
    return "0";
  }

  private static final long serialVersionUID = 29;

}


class ExtractorCWordSuff extends RareExtractor {

  private int num;

  public ExtractorCWordSuff(int num) {
    this.num = num;
  }

  String extract(History h, PairsHolder pH) {
    String word = TestSentence.toNice(pH.get(h, 0, false));
    if (word.length() < num) {
      return "impos";
    }
    return word.substring(word.length() - num, word.length());
  }

  private static final long serialVersionUID = 30;

  public String toString() {
    String cl = super.toString();
    return cl + " size " + num;
  }

}


class ExtractorCWordPref extends RareExtractor {

  private int num;

  public ExtractorCWordPref(int num) {
    this.num = num;
  }

  String extract(History h, PairsHolder pH) {
    String word = TestSentence.toNice(pH.get(h, 0, false));
    //word=word.toLowerCase();
    if (word.length() < num) {
      return "impos";
    } else {
      return word.substring(0, num);
      //return word.substring(word.length()-num,word.length());

    }
  }

  private static final long serialVersionUID = 31;

  public String toString() {
    String cl = super.toString();
    return cl + " size " + num;
  }

} // end class ExtractorCWordPref


class ExtractorsConjunction extends RareExtractor {

  private Extractor extractor1;
  private Extractor extractor2;

  public ExtractorsConjunction(Extractor e1, Extractor e2) {
    extractor1 = e1;
    extractor2 = e2;
  }

  String extract(History h, PairsHolder pH) {
    String ex1 = extractor1.extract(h);
    if (ex1.equals(zeroSt)) {
      return zeroSt;
    }
    String ex2 = extractor2.extract(h);
    if (ex2.equals(zeroSt)) {
      return zeroSt;
    }
    return ex1 + ":" + ex2;
  }

  private static final long serialVersionUID = 36;
}


class PluralAcronymDetector extends RareExtractor {

  public PluralAcronymDetector() {
  }

  private static boolean pluralAcronym(String s) {
    int len = s.length();
    len--;
    if (s.charAt(len) != 's') {
      return false;
    }
    for (int i = 0; i < len; i++) {
      if (!Character.isUpperCase(s.charAt(i))) {
        return false;
      }
    }
    return true;
  }

  String extract(History h, PairsHolder pH) {
    String s = pH.get(h, 0, false);
    if (pluralAcronym(s)) {
      return "1";
    }
    return "0";
  }

  private static final long serialVersionUID = 33;

}

class ctbPreDetector extends RareExtractor {
  private String t1;
  private int n1;
  public ctbPreDetector(String t2, int n2) {
    t1=t2;
    n1=n2;
  }

  String extract(History h, PairsHolder pH) {
    String s = TestSentence.toNice(pH.get(h,n1,false));

    if(!s.equals("") && CtbDict.getTagPre(t1, s.substring(0, 1)).equals("1"))
      return "1:"+t1;
    return "0:"+t1;
  }
  private static final long serialVersionUID = 43;
} // end class ctbPreDetector



class ctbSufDetector extends RareExtractor {
  private String t1;
  private int n1;
  public ctbSufDetector(String t2, int n2) {
    t1=t2;
    n1=n2;
  }

  String extract(History h, PairsHolder pH) {
    String s=TestSentence.toNice(pH.get(h,n1,false));

    if(!s.equals("") && CtbDict.getTagSuf(t1, s.substring(s.length()-1, s.length())).equals("1"))
      return "1:"+t1;
    return "0:"+t1;
  }
  private static final long serialVersionUID = 44;
} // end class ctbPreDetector



class ASBCPreDetector extends RareExtractor {
  private String t1;
  private int n1;
  public ASBCPreDetector(String t2, int n2) {
    t1=t2;
    n1=n2;
  }

  String extract(History h, PairsHolder pH) {
    String s=TestSentence.toNice(pH.get(h,n1,false));

    if(!s.equals("") && ASBCDict.getTagPre(t1, s.substring(0, 1)).equals("1"))
      return "1:"+t1;
    return "0:"+t1;
  }
  private static final long serialVersionUID = 53;
} // end class ASBCPreDetector



class ASBCSufDetector extends RareExtractor {
  private String t1;
  private int n1;
  public ASBCSufDetector(String t2, int n2) {
    t1=t2;
    n1=n2;
  }

  String extract(History h, PairsHolder pH) {
    String s=TestSentence.toNice(pH.get(h,n1,false));
    if(!s.equals("") && ASBCDict.getTagSuf(t1, s.substring(s.length()-1, s.length())).equals("1"))
      return "1:"+t1;
    return "0:"+t1;
  }
  private static final long serialVersionUID = 54;
} // end class ASBCPreDetector




class ASBCunkDetector extends RareExtractor {
  private String t1;
  private int n1;

  public ASBCunkDetector(String t2, int n2) {
    t1=t2;
    n1=n2;
  }

  String extract(History h, PairsHolder pH) {
    String s=TestSentence.toNice(pH.get(h,n1,false));

    if(ASBCunkDict.getTag(t1, s).equals("1"))
      return "1:"+t1;
    return "0:"+t1;
  }
  private static final long serialVersionUID = 57;
} // end class ASBCunkDetector

class CTBunkDictDetector extends RareExtractor {
  private String t1;
  private int n1;

  public CTBunkDictDetector(String t2, int n2) {
    t1=t2;
    n1=n2;
  }

  String extract(History h, PairsHolder pH) {
    String s=TestSentence.toNice(pH.get(h,n1,false));

    if(CTBunkDict.getTag(t1, s).equals("1"))
      return "1:"+t1;
    return "0:"+t1;
  }
  private static final long serialVersionUID = 80;
} // end class CTBunkDictDetector

class ExtractorWordShapeClassifier extends Extractor {

  private int wordShaper;

  public ExtractorWordShapeClassifier(int position, String wsc) {
    super(position, false);
    wordShaper = WordShapeClassifier.lookupShaper(wsc);
  }

  public static final String naTag = "NA";

  public boolean isPopulated(BinaryFeature f) {
    return f.indexedValues.length > GlobalHolder.rareWordMinFeatureThresh;
  }

  String extract(History h, PairsHolder pH) {
    String s = super.extract(h, pH);
    return WordShapeClassifier.wordShape(s, wordShaper, false);
  }

  private static final long serialVersionUID = 101L;

}


