//GlobalHolder -- StanfordMaxEnt, A Maximum Entropy Toolkit
//Copyright (c) 2002-2008 Leland Stanford Junior University


//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or (at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

//For more information, bug reports, fixes, contact:
//Christopher Manning
//Dept of Computer Science, Gates 1A
//Stanford CA 94305-9010
//USA
//    Support/Questions: java-nlp-user@lists.stanford.edu
//    Licensing: java-nlp-support@lists.stanford.edu
//http://www-nlp.stanford.edu/software/tagger.shtml

package edu.stanford.nlp.tagger.maxent;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import edu.stanford.nlp.io.InDataStreamFile;
import edu.stanford.nlp.io.OutDataStreamFile;
import edu.stanford.nlp.io.PrintFile;
import edu.stanford.nlp.maxent.iis.LambdaSolve;

/** This class holds many global variables and other things that are used by
 *  the Stanford MaxEnt Part-of-speech Tagger package.
 *
 *  @author Kristina Toutanova
 *  @author Anna Rafferty
 *  @version 1.1
 */
public class GlobalHolder {

  private GlobalHolder() {}

  static TaggerExperiments domain;
  static Dictionary dict = new Dictionary();
  static TTags tags;

  static byte[][] fnumArr;
  static LambdaSolveTagger prob;
  static HashSet<FeatureKey> sTemplates = new HashSet<FeatureKey>();
  static HashMap<FeatureKey,Integer> fAssociations = new HashMap<FeatureKey,Integer>();
  static HistoryTable tHistories = new HistoryTable();
  static TemplateHash tFeature = new TemplateHash();
  static PairsHolder pairs = new PairsHolder();
  static Extractors extractors;
  static Extractors extractorsRare;
  static AmbiguityClasses ambClasses = new AmbiguityClasses();
  static CollectionTaggerOutputs collectionTaggers = new CollectionTaggerOutputs(0);
  static boolean alltags = false;
  static HashMap<String, HashSet<String>> tagTokens = new HashMap<String, HashSet<String>>();

  /**
   * Determines which words are considered rare.  All words with count
   * in the training data less than this number (standardly, &lt; 5) are
   * considered rare.
   */
  static int rareWordThresh = 5;

  /**
   * Determines which features are included in the model.  The model
   * includes features that occurred strictly more times than this number
   * (standardly, &gt; 5) in the training data.  Here I look only at the
   * history (not the tag), so the history appearing this often is enough.
   */
  static int minFeatureThresh = 5;

  /**
   * This is a special threshold for the current word feature. Only words that have occurred &gt this number of times
   * will generate features with all of their occurring tags
   */
  static int curWordMinFeatureThresh = 2;

  /**
   * Determines which rare word features are included in the model.
   * The features for rare words have a strictly higher support than
   * this number.
   */
  static int rareWordMinFeatureThresh = 10;

  /**
   * If using equivalence classes on following words, words that occur
   * strictly more than this number of times (in total with any tag)
   * are sufficiently frequent to form an equivalence class
   * by themselves. (Not used unless using equivalence classes.)
   */
  static int veryCommonWordThresh = 250;


  static int xSize;
  static int ySize;
  static boolean occuringTagsOnly = false;

  static boolean initted = false;

  static final boolean VERBOSE = false;



  public static LambdaSolve getLambdaSolve() {
    return prob;
  }

  public static void init() {
    init(null);
  }

  public static void init(TaggerConfig config) {
    if (initted) return;

    extractors = new Extractors();
    extractorsRare = new Extractors();

    String lang, arch;
    String[] openClassTags, closedClassTags;

    if (config == null) {
      lang = "english";
      arch = "left3words";
      openClassTags = new String[0];
      closedClassTags = new String[0];
    } else {
      lang = config.getLang();
      arch = config.getArch();
      openClassTags = config.getOpenClassTags();
      closedClassTags = config.getClosedClassTags();
    }

    if (((openClassTags.length > 0) && !lang.equals("")) || ((closedClassTags.length > 0) && !lang.equals("")) || ((closedClassTags.length > 0) && (openClassTags.length > 0))) {
      throw new RuntimeException("At least two of lang (\"" + lang + "\"), openClassTags (length " + openClassTags.length + ": " + java.util.Arrays.toString(openClassTags) + ")," +
          "and closedClassTags (length " + closedClassTags.length + ": " + java.util.Arrays.toString(closedClassTags) + ") specified---you must choose one!");
    } else if ((openClassTags.length == 0) && lang.equals("") && (closedClassTags.length == 0) && !config.getLearnClosedClassTags()) {
      System.err.println("warning: no language set, no open-class tags specified, and no closed-class tags specified; assuming ALL tags are open class tags");
    }

    if (openClassTags.length > 0) {
      tags = new TTags();
      tags.setOpenClassTags(openClassTags);
    } else if(closedClassTags.length > 0) {
      tags = new TTags();
      tags.setClosedClassTags(closedClassTags);
    } else {
      tags = new TTags(lang);
    }


    if (config != null) {
      rareWordThresh = config.getRareWordThresh();
      minFeatureThresh = config.getMinFeatureThresh();
      curWordMinFeatureThresh = config.getCurWordMinFeatureThresh();
      rareWordMinFeatureThresh = config.getRareWordMinFeatureThresh();
      veryCommonWordThresh = config.getVeryCommonWordThresh();
    }

    if (arch.equalsIgnoreCase("left3words")) {
      extractors.init(ExtractorFrames.eFrames_left3words);
      extractorsRare.init(ExtractorFramesRare.getExtractorFramesRare("naacl2003", config));
    } else if (arch.equalsIgnoreCase("left3wordswordshapes")) {
      extractors.init(ExtractorFrames.eFrames_left3words);
      extractorsRare.init(ExtractorFramesRare.getExtractorFramesRare("naacl2003wordshape3", config));
    } else if (arch.equalsIgnoreCase("left5wordswordshapes")) {
      extractors.init(ExtractorFrames.eFrames_left5words);
      extractorsRare.init(ExtractorFramesRare.getExtractorFramesRare("naacl2003wordshape3", config));
    } else if (arch.equalsIgnoreCase("bidirectional")) {
      extractors.init(ExtractorFrames.eFrames_bidirectional);
      extractorsRare.init(ExtractorFramesRare.getExtractorFramesRare("naacl2003", config));
    } else if (arch.equalsIgnoreCase("bidirectionalwordshapes")) {
      extractors.init(ExtractorFrames.eFrames_bidirectional);
      extractorsRare.init(ExtractorFramesRare.getExtractorFramesRare("naacl2003wordshape3", config));
    } else if (arch.equalsIgnoreCase("bidirectionalwordshapes5words")) {
      extractors.init(ExtractorFrames.eFrames_bidirectional5words);
      extractorsRare.init(ExtractorFramesRare.getExtractorFramesRare("naacl2003wordshape3", config));
    } else if (arch.equalsIgnoreCase("sighan2005")) {
      if(config.getUseChineseDict()) {
        //First set up the dictionary prefix for the Chinese dictionaries
        ASBCDict.setPathPrefix(config.getChineseDictionaryPath());
        ASBCunkDict.setPathPrefix(config.getChineseDictionaryPath());
        CtbDict.setPathPrefix(config.getChineseDictionaryPath());
        CTBunkDict.setPathPrefix(config.getChineseDictionaryPath());
      }
      extractors.init(ExtractorFrames.eFrames_sighan2005);
      extractorsRare.init(ExtractorFramesRare.getExtractorFramesRare("sighan2005", config));
    } else if (arch.equalsIgnoreCase("generic")) {
      extractors.init(ExtractorFrames.eFrames_generic);
      extractorsRare.init(ExtractorFramesRare.getExtractorFramesRare("generic", config));
    } else if (arch.equalsIgnoreCase("1word")) {
      extractors.init(ExtractorFrames.eFrames_1word);
      extractorsRare.init(ExtractorFramesRare.getExtractorFramesRare("naacl2003", config));
    } else if (arch.equalsIgnoreCase("3words")) {
      extractors.init(ExtractorFrames.eFrames_3words);
      extractorsRare.init(ExtractorFramesRare.getExtractorFramesRare("naacl2003wordshape3", config));
    } else if (arch.equalsIgnoreCase("5words")) {
      extractors.init(ExtractorFrames.eFrames_5words);
      extractorsRare.init(ExtractorFramesRare.getExtractorFramesRare("naacl2003", config));
    } else if (arch.equalsIgnoreCase("7words")) {
      extractors.init(ExtractorFrames.eFrames_7words);
      extractorsRare.init(ExtractorFramesRare.getExtractorFramesRare("naacl2003", config));
    } else if (arch.equalsIgnoreCase("german")) {
      extractors.init(ExtractorFrames.eFrames_german);
      extractorsRare.init(ExtractorFramesRare.getExtractorFramesRare("generic", config));
    } else {
      throw new RuntimeException("arch must be one of [left3words, bidirectional, generic, 1word, 3words, 5words, sighan2005], not: " + arch);
    }

    initted = true;
  }

  /** Adds a FeatureKey to the set of known FeatureKeys.
   *
   * @param s The feature key to be added
   * @return Whether the key was already known (false) or added (true)
   */
  public static boolean add(FeatureKey s) {
    if ((sTemplates.contains(s))) {
      return false;
    }
    sTemplates.add(s);
    return true;
  }


  public static void putAssoc(FeatureKey s, int num) {
    fAssociations.put(s, Integer.valueOf(num));
  }


  public static int getNum(FeatureKey s) {
    return getNum(s, fAssociations);
  }


  public static int getNum(FeatureKey s, HashMap<FeatureKey, Integer> fAssocs) {
    Integer num = fAssocs.get(s);
    if (num == null) {
      return -1;
    } else {
      return num.intValue();
    }
  }


  /**
   * serialize the TaggerConfig
   */
  public static void saveConfig(TaggerConfig config,OutputStream os) throws Exception {
    ObjectOutputStream out = new ObjectOutputStream(os);
    out.writeObject(config);
  }

  /**
   * read in the TaggerConfig
   */
  public static TaggerConfig readConfig(InputStream file) throws Exception {
    ObjectInputStream in = new ObjectInputStream(file);
    TaggerConfig config = (TaggerConfig) in.readObject();
    return config;
  }


  /**
   * serialize the ExtractorFrames and ExtractorFramesRare in filename
   */
  public static void saveExtractors(OutputStream os) throws Exception {

    ObjectOutputStream out = new ObjectOutputStream(os);

    System.out.println(extractors.toString() + "\nrare" + extractorsRare.toString());
    out.writeObject(extractors);
    out.writeObject(extractorsRare);
    //out.close();
  }

  /**
   * read the extractors from a filename
   */
  public static void readExtractors(String filename) throws Exception {

    ObjectInputStream in = new ObjectInputStream(new FileInputStream(filename));
    readExtractors(in);
    in.close();
  }

  /**
   * read the extractors from a filename
   */
  public static void readExtractors(InputStream file) throws Exception {

    ObjectInputStream in = new ObjectInputStream(file);
    try {
      extractors = (Extractors) in.readObject();
      //System.err.println("Common word extractors: "+extractors.toString());

    } catch (Exception e) {
      e.printStackTrace();
    }
    extractorsRare = (Extractors) in.readObject();
    //System.err.println(extractorsRare.toString());
    //extractorsRare=new Extractors();
    //extractorsRare.init(ExtractorFramesRare.eFrames);
    int left = extractors.leftContext();
    int left_u = extractorsRare.leftContext();
    if (left_u > left) {
      left = left_u;
    }
    TestSentence.leftContext = left;
    int right = extractors.rightContext();
    int right_u = extractorsRare.rightContext();
    if (right_u > right) {
      right = right_u;
    }
    TestSentence.rightContext = right;
  }



  public static void save_prev(String filename) {
    try {
      OutDataStreamFile rf = new OutDataStreamFile(filename);
      save_prev(null, rf);
      rf.close();
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public static void save_prev(TaggerConfig config, OutDataStreamFile file) {
    try {
//      System.out.println("Tag counts are:");
//      for(String key : GlobalHolder.tagTokens.keySet()) {
//        System.out.println(key + "\t" + GlobalHolder.tagTokens.get(key).size());
//      }
      //config.writeConfig(file);
      saveConfig(config, file);
      file.writeInt(xSize);
      file.writeInt(ySize);
      dict.save(file);
      tags.save(file);

      saveExtractors(file);
      //write a boolean as to whether you're saving the config - this is messy and probably
      // not the best thing to do...
//      if(config != null) {
//        file.writeBoolean(true);
//        saveConfig(config, file);
//      } else {
//        file.writeBoolean(false);
//      }
      file.writeInt(fAssociations.size());
      Set<FeatureKey> set = fAssociations.keySet();
      for (FeatureKey fk: set) {
        int numF = fAssociations.get(fk).intValue();
        file.writeInt(numF);
        fk.save(file);

      }

      //we don't save these because we can reconstruct them on opening
      //TODO: test time/space tradeoff of saving versus not saving
      //ambClasses.save(rf);

    } catch (Exception e) {
      e.printStackTrace();
    }
  }


  /**
   * This reads the .assoc file.  It is only used by LambdaSolveTagger.java
   * The same associations also appear in the main file, and are read by
   * read(), read_prev().
   *
   * @param modelFilename The string .assoc is appended and feature
   *    associations are then read from this file
   * @return The feature associations HashMap, or null if there is an error
   */
  public static HashMap<FeatureKey, Integer> readAssociations(String modelFilename) {
    try {
      HashMap<FeatureKey, Integer> fAssocs = new HashMap<FeatureKey, Integer>();
      String filename = modelFilename + ".assoc";
      InDataStreamFile rf = new InDataStreamFile(filename);
      int sizeAssoc = rf.readInt();
      for (int i = 0; i < sizeAssoc; i++) {
        int numF = rf.readInt();
        FeatureKey fK = new FeatureKey();
        fK.read(rf);
        fAssocs.put(fK, Integer.valueOf(numF));
      }
      rf.close();
      return fAssocs;
    } catch (Exception e) {
      e.printStackTrace();
    }
    return null;
  }

  public static void release_mem() {

    // release dict
    dict.release();
    pairs.release();
    tHistories.release();
    tFeature.release();
    fAssociations.clear();
  }

  public static void save_after(OutDataStreamFile file) {
    try {
      prob.save(file);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /**
   * This method is provided for backwards compatibility with the old tagger.  It reads
   * a tagger that was saved as multiple files into the current format and saves it back
   * out as a single file, newFilename.
   * @param filename
   * @param newFilename
   * @return
   * @throws Exception
   */
  public static boolean convertMultifileTagger(String filename, String newFilename, TaggerConfig config) throws Exception {
    InDataStreamFile rf = new InDataStreamFile(filename);
    OutDataStreamFile of = new OutDataStreamFile(newFilename);
    GlobalHolder.init(config);
    if (VERBOSE) {
      System.err.println(" length of holder " + new File(filename).length());
    }

    xSize = rf.readInt();
    ySize = rf.readInt();
    dict.read(filename + ".dict");

    if (VERBOSE) {
      System.err.println(" dictionary read ");
    }
    tags.read(filename + ".tags");
    readExtractors(filename + ".ex");

    dict.setAmbClasses();

    int[] numFA = new int[extractors.getSize() + extractorsRare.getSize()];
    int sizeAssoc = rf.readInt();
    PrintFile pfVP = null;
    if (VERBOSE) {
      pfVP = new PrintFile("pairs.txt");
    }
    for (int i = 0; i < sizeAssoc; i++) {
      int numF = rf.readInt();
      FeatureKey fK = new FeatureKey();
      fK.read(rf);
      numFA[fK.num]++;
      fAssociations.put(fK, Integer.valueOf(numF));
    }

    save_prev(null, of);
    if (VERBOSE) {
      pfVP.close();
    }
    if (VERBOSE) {
      for (int k = 0; k < numFA.length; k++) {
        System.err.println(" Number of features of kind " + k + " " + numFA[k]);
      }
    }
    prob = new LambdaSolveTagger(filename + ".prob");
    if (VERBOSE) {
      System.err.println(" prob read ");
    }
    save_after(of);
    rf.close();
    of.close();
    return true;
  }


  public static boolean read(String filename) throws Exception {
    InDataStreamFile rf = new InDataStreamFile(filename);

    if (VERBOSE) {
      System.err.println(" length of holder " + new File(filename).length());
    }
    readConfig(rf);//we've actually already read and used the config

    xSize = rf.readInt();
    ySize = rf.readInt();
    dict.read(rf);

    if (VERBOSE) {
      System.err.println(" dictionary read ");
    }
    //if(tags == null) tags = new TTags();//this will just be written over anyway - seems like read should actually be static...
    tags.read(rf);
    readExtractors(rf);
//    boolean configSaved = rf.readBoolean();
//    if(configSaved)
//      readConfig(rf);
    dict.setAmbClasses();

    int[] numFA = new int[extractors.getSize() + extractorsRare.getSize()];
    int sizeAssoc = rf.readInt();
    PrintFile pfVP = null;
    if (VERBOSE) {
      pfVP = new PrintFile("pairs.txt");
    }
    for (int i = 0; i < sizeAssoc; i++) {
      int numF = rf.readInt();
      FeatureKey fK = new FeatureKey();
      fK.read(rf);
      numFA[fK.num]++;
      fAssociations.put(fK, Integer.valueOf(numF));
    }
    if (VERBOSE) {
      pfVP.close();
    }
    if (VERBOSE) {
      for (int k = 0; k < numFA.length; k++) {
        System.err.println(" Number of features of kind " + k + " " + numFA[k]);
      }
    }
    prob = new LambdaSolveTagger(rf);
    if (VERBOSE) {
      System.err.println(" prob read ");
    }
    rf.close();
    //tFile.close();
    return true;
  }



  public static boolean isRare(String word) {
    return (dict.sum(word) < rareWordThresh);
  }

}
