package edu.stanford.nlp.tagger.maxent;

import java.io.IOException;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Properties;

import edu.stanford.nlp.io.InDataStreamFile;
import edu.stanford.nlp.util.StringUtils;

/**
 * Reads and stores configuration information for a POS tagger.
 *
 *  @author William Morgan
 *  @author Anna Rafferty
 */
public class TaggerConfig extends Properties /* Inherits implementation of serializable! */ {
  private static final long serialVersionUID = -4136407850147157497L;

  private String[] openClassTags, closedClassTags, xmlInput;

  public enum Mode {
    TRAIN, TEST, TAG, CONVERT
  }
  private Mode mode = Mode.TRAIN;

  /* defaults */
  private static String
  SEARCH = "cg",
  DELIMITER = "/",
  TOKENIZE = "true",
  DEBUG = "false",
  ITERATIONS = "100",
  RARE_WORD_THRESH = String.valueOf(GlobalHolder.rareWordThresh),
  MIN_FEATURE_THRESH = String.valueOf(GlobalHolder.minFeatureThresh),
  CUR_WORD_MIN_FEATURE_THRESH = String.valueOf(GlobalHolder.curWordMinFeatureThresh),
  RARE_WORD_MIN_FEATURE_THRESH = String.valueOf(GlobalHolder.rareWordMinFeatureThresh),
  VERY_COMMON_WORD_THRESH = String.valueOf(GlobalHolder.veryCommonWordThresh),
  SIGMA_SQUARED = String.valueOf(0.5),
  ENCODING = "UTF-8",
  LEARN_CLOSED_CLASS = "false",
  CLOSED_CLASS_THRESHOLD = String.valueOf(TTags.CLOSED_TAG_THRESHOLD),
  USE_DIST_SIM = "false",
  DIST_SIM_PATH = "/u/nlp/data/pos_tags_are_useless/egw.bnc.200.pruned",
  CHINESE_DICT_PATH = "/u/nlp/data/pos-tagger/dictionary",
  USE_CHINESE_DICT = "false",
  VERBOSE = "false",
  SGML = "false",
  INIT_FROM_TREES = "false",
  LANG = "",
  TOKENIZER_FACTORY = "",
  XML_OUTPUT = "false";

  private static HashMap<String, String> defaultValues = new HashMap<String, String>();
  static {
    defaultValues.put("search", SEARCH);
    defaultValues.put("delimiter", DELIMITER);
    defaultValues.put("tokenize", TOKENIZE);
    defaultValues.put("debug", DEBUG);
    defaultValues.put("iterations", ITERATIONS);
    defaultValues.put("rareWordThresh", RARE_WORD_THRESH);
    defaultValues.put("minFeatureThresh", MIN_FEATURE_THRESH);
    defaultValues.put("curWordMinFeatureThresh", CUR_WORD_MIN_FEATURE_THRESH);
    defaultValues.put("rareWordMinFeatureThresh", RARE_WORD_MIN_FEATURE_THRESH);
    defaultValues.put("veryCommonWordThresh", VERY_COMMON_WORD_THRESH);
    defaultValues.put("sigmaSquared", SIGMA_SQUARED);
    defaultValues.put("encoding", ENCODING);
    defaultValues.put("learnClosedClassTags", LEARN_CLOSED_CLASS);
    defaultValues.put("closedClassTagThreshold", CLOSED_CLASS_THRESHOLD);
    defaultValues.put("useDistSim", USE_DIST_SIM);
    defaultValues.put("distSimPath", DIST_SIM_PATH);
    defaultValues.put("chineseDictionaryPath", CHINESE_DICT_PATH);
    defaultValues.put("useChineseDictionaries", USE_CHINESE_DICT);
    defaultValues.put("verbose", VERBOSE);
    defaultValues.put("sgml", SGML);
    defaultValues.put("initFromTrees", INIT_FROM_TREES);
    defaultValues.put("lang", LANG);
    defaultValues.put("tokenizerFactory", TOKENIZER_FACTORY);
    defaultValues.put("xmlOutput", XML_OUTPUT);
  }

  /**
   * This constructor just for creating an instance with default values.
   */
  private TaggerConfig() {
    super();
    this.putAll(defaultValues);
  }

  public TaggerConfig(String[] args) throws IOException {
    super();
    Properties props = StringUtils.argsToProperties(args);

    if (props.getProperty("") != null) {
      throw new RuntimeException("unknown argument(s): \"" + props.getProperty("") + "\"");
    }

    if (props.getProperty("genprops") != null) {
      printGenProps();
      System.exit(0);
    }

    //First thing to do is check if we're loading a classifier - if so, we need to load the config file from there
    //Figure out what mode we're in
    if(props.containsKey("trainFile")) {
      //Training mode
      mode = Mode.TRAIN;
      this.setProperty("file", props.getProperty("trainFile", "").trim());
    } else if(props.containsKey("testFile")) {
      //Testing mode
      mode = Mode.TEST;
      this.setProperty("file", props.getProperty("testFile", "").trim());
    } else if(props.containsKey("textFile")) {
      //Tagging mode
      mode = Mode.TAG;
      this.setProperty("file", props.getProperty("textFile", "").trim());
    } else if(props.containsKey("convertToSingleFile")) {
      mode = Mode.CONVERT;
      this.setProperty("file", props.getProperty("convertToSingleFile").trim());
    } else {
      mode = Mode.TAG;
      this.setProperty("file", "stdin");
    }
    //for any mode other than train, we load a classifier, which means we load a config - model always needs to be specified
    //on command line/in props file
    //Get the path to the model (or the path where you'd like to save the model); this is necessary for training, testing, and tagging
    this.setProperty("model", props.getProperty("model", "").trim());
    if (this.getProperty("model").equals("")) {
      throw new RuntimeException("'model' parameter must be specified");
    }

    /* Try and use the default properties from the model */
    //Properties modelProps = new Properties();
    TaggerConfig oldConfig = new TaggerConfig();
    if(mode != Mode.TRAIN) {
      try {
        InDataStreamFile in = new InDataStreamFile(this.getProperty("model"));
        //oldConfig = TaggerConfig.readConfig(in);
        oldConfig = GlobalHolder.readConfig(in);
        in.close();
        System.err.println("Trying to load default properties from trained tagger");
        //modelProps.load(new BufferedInputStream(new FileInputStream(mpfn)));
      } catch (Exception e) {
        System.err.println("No such trained tagger config file found.");
      }
    }



    this.setProperty("search", props.getProperty("search", oldConfig.getProperty("search")).trim().toLowerCase());
    if ( ! this.getProperty("search").equals("cg") && ! this.getProperty("search").equals("iis")) {
      throw new RuntimeException("'search' must be one of 'cg' or 'iis': " + this.getProperty("search"));
    }

    this.setProperty("sigmaSquared", props.getProperty("sigmaSquared", oldConfig.getProperty("sigmaSquared")));
    //sigmaSquared = Double.parseDouble(props.getProperty("sigmaSquared", (oldConfig.getSigmaSquared() == 0 ? defaultValues.get("sigmaSquared") : Double.toString(oldConfig.getSigmaSquared()))));

    this.setProperty("delimiter", props.getProperty("delimiter", oldConfig.getProperty("delimiter")));
    //delimiter = props.getProperty("delimiter", (oldConfig.getDelimiter() == null ? defaultValues.get("delimiter") : oldConfig.getDelimiter()));
    TestSentence.delimiter = ((this.getDelimiter() == null || this.getDelimiter().equals("")) ? TestSentence.delimiter : this.getDelimiter());


    this.setProperty("iterations", props.getProperty("iterations", oldConfig.getProperty("iterations")));
    //iterations = Integer.parseInt(props.getProperty("iterations", (oldConfig.getIterations() == 0 ? defaultValues.get("iterations") : Integer.toString(oldConfig.getIterations()))));
    this.setProperty("rareWordThresh", props.getProperty("rareWordThresh", oldConfig.getProperty("minFeatureThresh")));
    //rareWordThresh = Integer.parseInt(props.getProperty("rareWordThresh", (oldConfig.getRareWordThresh() == 0 ? defaultValues.get("rareWordThresh") : Integer.toString(oldConfig.getRareWordThresh()))));
    this.setProperty("minFeatureThresh", props.getProperty("minFeatureThresh", oldConfig.getProperty("minFeatureThresh")));
    //minFeatureThresh = Integer.parseInt(props.getProperty("minFeatureThresh", (oldConfig.getMinFeatureThresh() == 0 ? defaultValues.get("minFeatureThresh") : Integer.toString(oldConfig.getMinFeatureThresh()))));
    this.setProperty("curWordMinFeatureThresh", props.getProperty("curWordMinFeatureThresh", oldConfig.getProperty("curWordMinFeatureThresh")));
    //curWordMinFeatureThresh = Integer.parseInt(props.getProperty("curWordMinFeatureThresh", (oldConfig.getCurWordMinFeatureThresh() == 0 ? defaultValues.get("curWordMinFeatureThresh") : Integer.toString(oldConfig.getCurWordMinFeatureThresh()))));
    this.setProperty("rareWordMinFeatureThresh", props.getProperty("rareWordMinFeatureThresh", oldConfig.getProperty("rareWordMinFeatureThresh")));
    //rareWordMinFeatureThresh = Integer.parseInt(props.getProperty("rareWordMinFeatureThresh", (oldConfig.getRareWordMinFeatureThresh() == 0 ? defaultValues.get("rareWordMinFeatureThresh") : Integer.toString(oldConfig.getRareWordMinFeatureThresh()))));
    this.setProperty("veryCommonWordThresh", props.getProperty("veryCommonWordThresh", oldConfig.getProperty("veryCommonWordThresh")));
    //veryCommonWordThresh = Integer.parseInt(props.getProperty("veryCommonWordThresh", (oldConfig.getVeryCommonWordThresh() == 0 ? defaultValues.get("veryCommonWordThresh") : Integer.toString(oldConfig.getVeryCommonWordThresh()))));


    this.setProperty("lang", props.getProperty("lang", oldConfig.getProperty("lang")));
    //lang = props.getProperty("lang", (oldConfig.getLang() == null ? defaultValues.get("lang") : oldConfig.getLang()));

    //this is a property that is stored (not like the general properties)
    String ocs = props.getProperty("openClassTags");
    if((ocs == null) || ocs.equals("") || ocs.equals("true"))  {/* necessary because of peculiarities of java -what does this note mean?? i seriously doubt its veracity...*/
      if(oldConfig.getOpenClassTags() == null)
        openClassTags = new String[0];
      else
        openClassTags = oldConfig.getOpenClassTags();
    } else
      openClassTags = ocs.split("\\s+");

    //this is a property that is stored (not like the general properties)
    String ccs = props.getProperty("closedClassTags");

    if((ccs == null) || ccs.equals("") || ccs.equals("true")) {/* necessary because of peculiarities of java */
      if(oldConfig.getClosedClassTags() == null)
        closedClassTags = new String[0];
      else
        closedClassTags = oldConfig.getClosedClassTags();
    } else
      closedClassTags = ccs.split("\\s+");

    this.setProperty("learnClosedClassTags", props.getProperty("learnClosedClassTags", oldConfig.getProperty("learnClosedClassTags")));
    //learnClosedClassTags = Boolean.parseBoolean(props.getProperty("learnClosedClassTags", Boolean.toString(oldConfig.getLearnClosedClassTags())));


    this.setProperty("closedClassTagThreshold", props.getProperty("closedClassTagThreshold", oldConfig.getProperty("closedClassTagThreshold")));
    //closedClassTagThreshold = Integer.parseInt(props.getProperty("closedClassTagThreshold", (oldConfig.getClosedTagThreshold() == 0 ? defaultValues.get("closedClassTagThreshold") : Integer.toString(oldConfig.getClosedTagThreshold()))));

    this.setProperty("arch", props.getProperty("arch", oldConfig.getProperty("arch")));
    //arch = props.getProperty("arch", (oldConfig.getArch() == null ? defaultValues.get("arch") : oldConfig.getArch()));
    this.setProperty("tokenize", props.getProperty("tokenize", oldConfig.getProperty("tokenize")));
    //tokenize = Boolean.parseBoolean(props.getProperty("tokenize", Boolean.toString(oldConfig.getLearnClosedClassTags())));
    this.setProperty("tokenizerFactory", props.getProperty("tokenizerFactory", oldConfig.getProperty("tokenizerFactory")));
    //tokenizerFactory = props.getProperty("tokenizerFactory", (oldConfig.getTokenizerFactory() == null ? defaultValues.get("tokenizerFactory") : oldConfig.getTokenizerFactory()));

    this.setProperty("debugPrefix", props.getProperty("debugPrefix", oldConfig.getProperty("debugPrefix", "")));
    //debugPrefix = props.getProperty("debugPrefix", "");
    this.setProperty("debug", props.getProperty("debug", DEBUG));
    //debug = Boolean.parseBoolean(props.getProperty("debug", DEBUG));

    this.setProperty("encoding", props.getProperty("encoding", oldConfig.getProperty("encoding")));
    //encoding = props.getProperty("encoding", (oldConfig.getEncoding() == null ? defaultValues.get("encoding") : oldConfig.getEncoding()));
    this.setProperty("useDistSim", props.getProperty("useDistSim", oldConfig.getProperty("useDistSim")));
    //useDistSim = Boolean.parseBoolean(props.getProperty("useDistSim", modelProps.getProperty("useDistSim", USE_DIST_SIM)));
    this.setProperty("distSimPath", props.getProperty("distSimPath", oldConfig.getProperty("distSimPath")));
    //distSimPath = props.getProperty("distSimPath", modelProps.getProperty("distSimPath", DIST_SIM_PATH));
    this.setProperty("sgml", props.getProperty("sgml", oldConfig.getProperty("sgml")));
    //sgml = Boolean.parseBoolean(props.getProperty("sgml", modelProps.getProperty("sgml", SGML)));
    this.setProperty("verbose", props.getProperty("verbose", oldConfig.getProperty("verbose")));
    //verbose = Boolean.parseBoolean(props.getProperty("verbose", modelProps.getProperty("verbose", VERBOSE)));
    this.setProperty("initFromTrees", props.getProperty("initFromTrees", oldConfig.getProperty("initFromTrees")));
    //initFromTrees = Boolean.parseBoolean(props.getProperty("initFromTrees", modelProps.getProperty("initFromTrees", INIT_FROM_TREES)));

    //this is a property that is stored (not like the general properties)
    String xmlInputProp = props.getProperty("xmlInput", null);
    if((xmlInputProp == null) || xmlInputProp.equals("")) {/* necessary because of peculiarities of java */
      if(oldConfig.getXMLInput() == null)
        xmlInput = new String[0];
      else
        xmlInput = oldConfig.getXMLInput();
    } else
      xmlInput = xmlInputProp.split("\\s+");
    this.setProperty("xmlOutput", props.getProperty("xmlOutput", oldConfig.getProperty("xmlOutput")));
    //xmlOutput = props.getProperty("xmlOutput", modelProps.getProperty("xmlOutput", ""));


    //Chinese properties
    this.setProperty("chineseDictionaryPath", props.getProperty("chineseDictionaryPath", oldConfig.getProperty("chineseDictionaryPath")).trim());
    //chineseDictionaryPath = props.getProperty("chineseDictionaryPath",modelProps.getProperty("chineseDictionaryPath", CHINESE_DICT_PATH)).trim();
    this.setProperty("useChineseDictionaries", props.getProperty("useChineseDictionaries", oldConfig.getProperty("useChineseDictionaries")));
    //useChineseDict = Boolean.parseBoolean(props.getProperty("useChineseDictionaries", modelProps.getProperty("useChineseDictionaries", "false")));

  }

//public void writeConfig(OutDataStreamFile out) throws Exception {
////model
//out.writeUTF(model);

////search method (is this really necessary to save?)
//out.writeUTF(search);
////sigma squared (is this really necessary to save?)
//out.writeDouble(sigmaSquared);
////num iterations(is this really necessary to save?)
//out.writeInt(iterations);
////delimiter
//out.writeUTF(delimiter);
////rare word threshold (is this really necessary to save?)
//out.writeInt(rareWordThresh);
////min feature (is this really necessary to save?)
//out.writeInt(minFeatureThresh);
////cur word min feature threshold (is this really necessary to save?)
//out.writeInt(curWordMinFeatureThresh);
////rare word min feature threshold (is this really necessary to save?)
//out.writeInt(rareWordMinFeatureThresh);
////very common word threshold (is this really necessary to save?)
//out.writeInt(veryCommonWordThresh);
////lang (is this really necessary to save?)
//out.writeUTF(lang);
////we don't save the closed/open tag features because these are used only for training, not testing/tagging

////arch
//out.writeUTF(arch);
////tokenize
//out.writeBoolean(tokenize);
////tokenizer factory
//out.writeUTF(tokenizerFactory);
////we don't save debug b/c this isn't really a property of the tagger

////encoding
//out.writeUTF(encoding);
////useDistSim
//out.writeBoolean(useDistSim);
////distSimPath
//out.writeUTF(distSimPath);
////sgml
//out.writeBoolean(sgml);
////we don't save verbose b/c this isn't really a property of the tagger

////initFromTrees (not sure if this should be saved)
//out.writeBoolean(initFromTrees);
////xmlInput
//out.writeInt(xmlInput.length);
//for(int i = 0; i < xmlInput.length; i++) {
//out.writeUTF(xmlInput[i]);
//}

////xmlOutput - we don't save this b/c it's instance specific

////useChineseDict
//out.writeBoolean(useChineseDict);
//// chineseDictionaryPath
//out.writeUTF(chineseDictionaryPath);


//}

//public static TaggerConfig readConfig(InDataStreamFile in) throws Exception {
//System.out.println("Reading config");
//TaggerConfig config = new TaggerConfig();
////model
//config.model = in.readUTF();
////search method (is this really necessary to save?)
//config.search = in.readUTF();
////sigma squared (is this really necessary to save?)
//config.sigmaSquared = in.readDouble();
////num iterations(is this really necessary to save?)
//config.iterations = in.readInt();
////delimiter
//config.delimiter = in.readUTF();
////rare word threshold (is this really necessary to save?)
//config.rareWordThresh = in.readInt();
////min feature (is this really necessary to save?)
//config.minFeatureThresh = in.readInt();
////cur word min feature threshold (is this really necessary to save?)
//config.curWordMinFeatureThresh = in.readInt();
////rare word min feature threshold (is this really necessary to save?)
//config.rareWordMinFeatureThresh = in.readInt();
////very common word threshold (is this really necessary to save?)
//config.veryCommonWordThresh = in.readInt();
////lang (is this really necessary to save?)
//config.lang = in.readUTF();
////we don't save the closed/open tag features because these are used only for training, not testing/tagging

////arch
//config.arch = in.readUTF();
////tokenize
//config.tokenize = in.readBoolean();
////tokenizer factory
//config.tokenizerFactory = in.readUTF();
////we don't save debug b/c this isn't really a property of the tagger

////encoding
//config.encoding = in.readUTF();
////useDistSim
//config.useDistSim = in.readBoolean();
////distSimPath
//config.distSimPath = in.readUTF();
////sgml
//config.sgml = in.readBoolean();
////we don't save verbose b/c this isn't really a property of the tagger

////initFromTrees (not sure if this should be saved)
//config.initFromTrees = in.readBoolean();
////xmlInput
//int numFields = in.readInt();
//config.xmlInput = new String[numFields];
//for(int i = 0; i < numFields; i++) {
//config.xmlInput[i] = in.readUTF();
//}
////xmlOutput - we don't save this b/c it's instance specific

////useChineseDict
//config.useChineseDict = in.readBoolean();

//// chineseDictionaryPath
//config.chineseDictionaryPath = in.readUTF();

//return config;
//}


  public String getModel() { return getProperty("model"); }

  public String getFile() { return getProperty("file"); }

  public String getSearch() { return getProperty("search"); }

  public double getSigmaSquared() { return Double.parseDouble(getProperty("sigmaSquared")); }

  public int getIterations() { return Integer.parseInt(getProperty("iterations")); }

  public int getRareWordThresh() { return Integer.parseInt(getProperty("rareWordThresh")); }

  public int getMinFeatureThresh() { return Integer.parseInt(getProperty("minFeatureThresh")); }

  public int getCurWordMinFeatureThresh() { return Integer.parseInt(getProperty("curWordMinFeatureThresh")); }

  public int getRareWordMinFeatureThresh() { return Integer.parseInt(getProperty("rareWordMinFeatureThresh")); }

  public int getVeryCommonWordThresh() { return Integer.parseInt(getProperty("veryCommonWordThresh")); }

  public String getLang() { return getProperty("lang"); }

  public String[] getOpenClassTags() { return openClassTags; }

  public String[] getClosedClassTags() { return closedClassTags; }

  public boolean getLearnClosedClassTags() { return Boolean.parseBoolean(getProperty("learnClosedClassTags")); }

  public int getClosedTagThreshold() { return Integer.parseInt(getProperty("closedClassTagThreshold")); }

  public String getArch() { return getProperty("arch"); }

  public boolean getDebug() { return Boolean.parseBoolean(getProperty("debug")); }

  public String getDebugPrefix() { return getProperty("debugPrefix"); }

  public String getTokenizerFactory() { return getProperty("tokenizerFactory"); }

  public String getDelimiter() { return getProperty("delimiter"); }

  public boolean getTokenize() { return Boolean.parseBoolean(getProperty("tokenize")); }

  public String getEncoding() { return getProperty("encoding"); }

  public boolean getUseDistSim() { return Boolean.parseBoolean(getProperty("useDistSim")); }

  public String getDistSimPath() { return getProperty("distSimPath"); }

  public String getChineseDictionaryPath() { return getProperty("chineseDictionaryPath"); }

  public boolean getUseChineseDict() { return Boolean.parseBoolean(getProperty("useChineseDict")); }

  public String[] getXMLInput() { return xmlInput; }

  public String getXMLOutput() { return getProperty("xmlOutput"); }

  public boolean getInitFromTrees() { return Boolean.parseBoolean(getProperty("initFromTrees")); }

  public boolean getVerbose() { return Boolean.parseBoolean(getProperty("verbose")); }

  public boolean getSGML() { return Boolean.parseBoolean(getProperty("sgml")); }

  public void dump() { dump(System.err); }

  public void dump(PrintStream stream) {
    stream.println("                   model = " + getProperty("model"));
    if(mode == Mode.TRAIN)
      stream.println("                    trainFile = " + getProperty("file"));
    if(mode == Mode.TAG)
      stream.println("                    textFile = " + getProperty("file"));
    if(mode == Mode.TEST)
      stream.println("                    testFile = " + getProperty("file"));
    stream.println("               delimiter = " + getProperty("delimiter"));
    stream.println("        tokenizerFactory = " + getProperty("tokenizerFactory"));
    stream.println("                  search = " + getProperty("search"));
    stream.println("            sigmaSquared = " + getProperty("sigmaSquared"));
    stream.println("              iterations = " + getProperty("iterations"));
    stream.println("                    lang = " + getProperty("lang"));
    stream.println("           openClassTags = " + StringUtils.join(openClassTags, " "));
    stream.println("         closedClassTags = " + StringUtils.join(closedClassTags, " "));
    stream.println("    learnClosedClassTags = " + getProperty("learnClosedClassTags"));
    stream.println(" closedClassTagThreshold = " + getProperty("closedClassTagThreshold"));

    stream.println("                    arch = " + getProperty("arch"));
    stream.println("                   debug = " + getProperty("debug"));
    stream.println("             debugPrefix = " + getProperty("debugPrefix"));
    stream.println("                encoding = " + getProperty("encoding"));
    stream.println("                tokenize = " + getProperty("tokenize"));
    stream.println("          rareWordThresh = " + getProperty("rareWordThresh"));
    stream.println("        minFeatureThresh = " + getProperty("minFeatureThresh"));
    stream.println(" curWordMinFeatureThresh = " + getProperty("curWordMinFeatureThresh"));
    stream.println("rareWordMinFeatureThresh = " + getProperty("rareWordMinFeatureThresh"));
    stream.println("    veryCommonWordThresh = " + getProperty("veryCommonWordThresh"));
    stream.println("              useDistSim = " + getProperty("useDistSim"));
    stream.println("             distSimPath = " + getProperty("distSimPath"));



  }

  /**
   * Prints out the automatically generated props file - in its own
   * method to make code above easier to read
   */
  private static void printGenProps() {
    System.out.println("## Sample properties file for maxent tagger. This file is used for three main");
    System.out.println("## operations: training, testing, and tagging. It may also be used to convert a");
    System.out.println("## To train or test a model, or to tag something, run:");
    System.out.println("##   java edu.stanford.nlp.tagger.maxent.MaxentTagger -prop <properties file>");

    System.out.println("## Arguments can be overridden on the commandline, e.g.:");
    System.out.println("##   java ....MaxentTagger -prop <properties file> -file /other/file ");
    System.out.println("");
    System.out.println("# Model file name (created at train time; used at tag and test time)");
    System.out.println("# (you can leave this blank and specify it on the commandline with -model)");
    System.out.println("# model = ");
    System.out.println("");

    System.out.println("# Path to file to be operated on (trained from, tested against, or tagged)");
    System.out.println("# Specify -textFile <filename> to tag text in the given file, -trainFile <filename> to");
    System.out.println("# to train a model using data in the given file, or -testFile <filname> to test your");
    System.out.println("# model using data in the given file.  Alternatively, you may specify");
    System.out.println("# -convertToSingleFile <filename> to save an old, multi-file model (specified at -model) to");
    System.out.println("# the new single file format.  The new model will be saved in the file filename. If you choose to");
    System.out.println("# convert an old file, you must specify the correct 'arch' parameter used to create the original model.");
    System.out.println("# trainFile = ");
    System.out.println("");

    System.out.println("# Delimiter character that separates word and pos tags (for both training");
    System.out.println("# and test data).");
    System.out.println("# delimiter = " + DELIMITER);
    System.out.println("");

    System.out.println("# Encoding format in which files are stored.  If left blank, UTF-8 is assumed.");
    System.out.println("# encoding = " + ENCODING);
    System.out.println("");


    System.out.println("######### parameters for tag and test operations #########");
    System.out.println("");
    System.out.println("# Class to use for tokenization. Default blank value means Penn Treebank");
    System.out.println("# tokenization.  If you'd like to just assume that tokenization has been done,");
    System.out.println("# and the input is whitespace-tokenized, use");
    System.out.println("# edu.stanford.nlp.process.WhitespaceTokenizer or set tokenize to false.");
    System.out.println("# tokenizerFactory = ");
    System.out.println("");

    System.out.println("# Whether to tokenize text for tag and test operations. Default is true.");
    System.out.println("# If false, your text must already be whitespace tokenized.");
    System.out.println("# tokenize = " + TOKENIZE);
    System.out.println("");

    System.out.println("# Write debugging information (words, top words, unknown words). Useful for");
    System.out.println("# error analysis. Default is false.");
    System.out.println("# debug = "+ DEBUG);
    System.out.println("");

    System.out.println("# Prefix for debugging output (if debug == true). Default is to use the");
    System.out.println("# filename from 'file'");
    System.out.println("# debugPrefix = ");
    System.out.println("");

    System.out.println("######### parameters for training  #########");
    System.out.println("");
    System.out.println("# model architecture:");
    System.out.println("# 'left3words', 'left3wordswordshapes', left5wordswordshapes', 'bidirectional', ");
    System.out.println("# 'bidirectionalwordshapes', 'bidirectionalwordshapes5words', 'generic', '[1357]words?', or 'sighan2005'.");
    System.out.println("# This determines the features extracted.  'generic' is language independent.");
    System.out.println("# arch = ");
    System.out.println("");

    System.out.println("# language (for list of open-class tags). currently we have 'english', 'arabic', 'german', 'chinese'");
    System.out.println("# or 'polish' predefined. alternatively, you can specify openClassTags or closedClassTags");
    System.out.println("# below (only ONE of these three options may be specified). 'english' means penn treebank tags. 'polish' means");
    System.out.println("# some tags that some guy on the internet once used. See TTags class for more information; 'english' is the");
    System.out.println("# most accurate of the predefined languages ('arabic' and 'german' in particular are not very well defined).");
    System.out.println("# lang = ");
    System.out.println("");

    System.out.println("# a space-delimited list of open-class parts of speech");
    System.out.println("# alternatively, you can specify language above to use a pre-defined list or specify the closed class tags (below)");
    System.out.println("# openClassTags = ");
    System.out.println("");

    System.out.println("# a space-delimited list of closed-class parts of speech");
    System.out.println("# alternatively, you can specify language above to use a pre-defined list or specify the open class tags (above)");
    System.out.println("# closedClassTags = ");
    System.out.println("");

    System.out.println("# A boolean indicating whether you would like the trained model to set POS tags as closed");
    System.out.println("# based on their frequency in training; default is false.  The frequency threshold can be set below. ");
    System.out.println("# This option is ignored if any of {openClassTags, closedClassTags, lang} are specified.");
    System.out.println("# learnClosedClassTags = ");
    System.out.println("");

    System.out.println("# Used only if learnClosedClassTags=true.  Tags that have fewer tokens than this threshold are");
    System.out.println("# considered closed in the trained model.");
    System.out.println("# closedClassTagThreshold = ");
    System.out.println("");

    System.out.println("# search method. 'cg' (conjugate gradient (default)) or 'iis' (improved ");
    System.out.println("# iterative scaling)");
    System.out.println("# search = " + SEARCH);
    System.out.println("");

    System.out.println("# for conjugate gradient search, sigma-squared smoothing/regularization");
    System.out.println("# parameter. if left blank, the default is 0.5, which is usually okay");
    System.out.println("# sigmaSquared = " + SIGMA_SQUARED);
    System.out.println("");

    System.out.println("# for improved iterative scaling, the number of iterations");
    System.out.println("# iterations = " + ITERATIONS);
    System.out.println("");

    System.out.println("# rare word threshold. words that occur less than this number of");
    System.out.println("# times are considered rare words.");
    System.out.println("# rareWordThresh = " + RARE_WORD_THRESH);
    System.out.println("");

    System.out.println("# minimum feature threshold. features whose history appears less");
    System.out.println("# than this number of times are ignored.");
    System.out.println("# minFeatureThresh = " + MIN_FEATURE_THRESH);
    System.out.println("");

    System.out.println("# current word feature threshold. words that occur more than this");
    System.out.println("# number of times will generate features with all of their occurring");
    System.out.println("# tags.");
    System.out.println("# curWordMinFeatureThresh = " + CUR_WORD_MIN_FEATURE_THRESH);
    System.out.println("");

    System.out.println("# rare word minimum feature threshold. features of rare words whose histories");
    System.out.println("# appear less than this times will be ignored.");
    System.out.println("# rareWordMinFeatureThresh = " + RARE_WORD_MIN_FEATURE_THRESH);
    System.out.println("");

    System.out.println("# very common word threshold. words that occur more than this number of");
    System.out.println("# times will form an equivalence class by themselves. ignored unless");
    System.out.println("# you are using equivalence classes.");
    System.out.println("# veryCommonWordThresh = " + VERY_COMMON_WORD_THRESH);

    System.out.println("# useDistSim: Distributional similarity classes can be an added source of information");
    System.out.println("# about your words. An English distsim file is included, or you can use your own.  Distsim");
    System.out.println("# feature will be added in addition to your other features.");
    System.out.println("# useDistSim = " + USE_DIST_SIM);

    System.out.println("# distSimPath: Path to the distributional similarity file.  The included English distsim file");
    System.out.println("# is egw.bnc.200.pruned, located in the main part of speech tagger directory.");
    System.out.println("# distSimPath = ");

  }

  public Mode getMode() {
    return mode;
  }

}
